# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
import socket

from ..compat import mock
from ..fixtures.mocks import MockObject
from .base import CommandTestCase
from gandi.cli.commands import vm
from gandi.cli.core.base import GandiContextHelper


class VmTestCase(CommandTestCase):

    mocks = [('gandi.cli.core.base.GandiModule.exec_output',
              MockObject.exec_output)]

    def test_list(self):

        result = self.invoke_with_exceptions(vm.list, [])

        self.assertEqual(result.output, """hostname  : vm1426759833
state     : running
----------
hostname  : vm1426759844
state     : running
----------
hostname  : server01
state     : running
----------
hostname  : server02
state     : halted
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_id(self):

        result = self.invoke_with_exceptions(vm.list, ['--id'])

        self.assertEqual(result.output, """hostname  : vm1426759833
state     : running
id        : 152966
----------
hostname  : vm1426759844
state     : running
id        : 152964
----------
hostname  : server01
state     : running
id        : 152967
----------
hostname  : server02
state     : halted
id        : 152968
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_filter_state(self):

        result = self.invoke_with_exceptions(vm.list, ['--state', 'halted'])

        self.assertEqual(result.output, """hostname  : server02
state     : halted
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_filter_datacenter(self):

        result = self.invoke_with_exceptions(vm.list, ['--datacenter', 'FR'])

        self.assertEqual(result.output, """hostname  : server01
state     : running
----------
hostname  : server02
state     : halted
""")
        self.assertEqual(result.exit_code, 0)

    def test_info_ko_resource(self):

        result = self.invoke_with_exceptions(vm.info, [])
        self.assertEqual(result.exit_code, 2)

    def test_info_ok_one_resource(self):

        result = self.invoke_with_exceptions(vm.info, ['server01'])

        self.assertEqual(result.output, """hostname      : server01
state         : running
cores         : 1
memory        : 256
console       :
datacenter    : FR-SD2
----------
bandwidth     : 102400.0
ip4           : 95.142.160.181
ip6           : 2001:4b98:dc0:47:216:3eff:feb2:3862

label         : Debian 7 64 bits (HVM)
kernel_version: 3.12-x86_64 (hvm)
name          : sys_server01
size          : 3072
""")
        self.assertEqual(result.exit_code, 0)

    def test_info_ok_multiple_resources(self):
        args = ['server01', 'vm1426759833']
        result = self.invoke_with_exceptions(vm.info, args)

        self.assertEqual(result.output, """hostname      : server01
state         : running
cores         : 1
memory        : 256
console       :
datacenter    : FR-SD2
----------
bandwidth     : 102400.0
ip4           : 95.142.160.181
ip6           : 2001:4b98:dc0:47:216:3eff:feb2:3862

label         : Debian 7 64 bits (HVM)
kernel_version: 3.12-x86_64 (hvm)
name          : sys_server01
size          : 3072
----------
hostname      : vm1426759833
state         : running
cores         : 1
memory        : 256
console       :
datacenter    : LU-BI1
----------
bandwidth     : 102400.0
ip6           : 2001:4b98:dc2:43:216:3eff:fece:e25f

label         : Debian 7 64 bits (HVM)
kernel_version: 3.12-x86_64 (hvm)
name          : sys_1426759833
size          : 3072
""")
        self.assertEqual(result.exit_code, 0)

    def test_info_stat(self):

        result = self.invoke_with_exceptions(vm.info, ['server01', '--stat'])

        expected = u"""hostname      : server01
state         : running
cores         : 1
memory        : 256
console       :
datacenter    : FR-SD2
----------
bandwidth     : 102400.0
ip4           : 95.142.160.181
ip6           : 2001:4b98:dc0:47:216:3eff:feb2:3862

label         : Debian 7 64 bits (HVM)
kernel_version: 3.12-x86_64 (hvm)
name          : sys_server01
size          : 3072

vm network stats
in            :    ▁▁ ▂▁ ▁▁▁▂▂▁▁     ▉▃▁
out           :    ▁▃ ▉▂ ▆▆▁▃▄▁▁ ▁  ▁▉▅▇
disk network stats
read          :     ▁ ▁▁ ▁  ▉        ▁▃▁
write         :     ▁    ▁  ▂        ▉▁▁
"""
        self.assertEqual(result.output, expected)
        self.assertEqual(result.exit_code, 0)

    def test_datacenters(self):

        result = self.invoke_with_exceptions(vm.datacenters, [])

        self.assertEqual(result.output, """\
iso       : FR
name      : Equinix Paris
country   : France
dc_code   : FR-SD2
closing on: 25/12/2017
----------
iso       : US
name      : Level3 Baltimore
country   : United States of America
dc_code   : US-BA1
closing on: 25/12/2016
closed for: vm, paas
----------
iso       : LU
name      : Bissen
country   : Luxembourg
dc_code   : LU-BI1
----------
iso       : FR
name      : France, Paris
country   : France
dc_code   : FR-SD3
closed for: paas
----------
iso       : FR
name      : France, Paris
country   : France
dc_code   : FR-SD5
closed for: paas
""")
        self.assertEqual(result.exit_code, 0)

    def test_datacenters_id(self):

        result = self.invoke_with_exceptions(vm.datacenters, ['--id'])

        self.assertEqual(result.output, """\
iso       : FR
name      : Equinix Paris
country   : France
dc_code   : FR-SD2
id        : 1
closing on: 25/12/2017
----------
iso       : US
name      : Level3 Baltimore
country   : United States of America
dc_code   : US-BA1
id        : 2
closing on: 25/12/2016
closed for: vm, paas
----------
iso       : LU
name      : Bissen
country   : Luxembourg
dc_code   : LU-BI1
id        : 3
----------
iso       : FR
name      : France, Paris
country   : France
dc_code   : FR-SD3
id        : 4
closed for: paas
----------
iso       : FR
name      : France, Paris
country   : France
dc_code   : FR-SD5
id        : 5
closed for: paas
""")
        self.assertEqual(result.exit_code, 0)

    def test_kernels(self):

        result = self.invoke_with_exceptions(vm.kernels, [])

        linux_hvm_out = """\
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
version       : grub
version       : raw
"""
        linux_out_dc1 = """\
flavor        : linux
version       : 2.6.18 (deprecated)
version       : 2.6.27-compat-sysfs (deprecated)
version       : 2.6.32
version       : 2.6.27 (deprecated)
version       : 2.6.32-x86_64
version       : 2.6.36 (deprecated)
version       : 2.6.32-x86_64-grsec
version       : 2.6.36-x86_64 (deprecated)
version       : 3.2-i386
version       : 3.2-x86_64
version       : 3.2-x86_64-grsec
version       : 3.10-x86_64
version       : 3.10-i386
"""
        linux_out_dc2 = """\
flavor        : linux
version       : 2.6.18 (deprecated)
version       : 2.6.27-compat-sysfs (deprecated)
version       : 2.6.32
version       : 2.6.27 (deprecated)
version       : 2.6.32-x86_64
version       : 2.6.36 (deprecated)
version       : 2.6.32-x86_64-grsec
version       : 2.6.36-x86_64 (deprecated)
version       : 3.2-i386
version       : 3.2-x86_64
version       : 3.2-x86_64-grsec
version       : 3.10-x86_64
version       : 3.10-i386
"""
        linux_out_dc3 = """\
flavor        : linux
version       : 2.6.32
version       : 2.6.27 (deprecated)
version       : 2.6.32-x86_64
version       : 2.6.32-x86_64-grsec
version       : 3.2-i386
version       : 3.2-x86_64
version       : 3.2-x86_64-grsec
version       : 3.10-x86_64
version       : 3.10-i386
"""
        self.assertTrue(linux_hvm_out in result.output)
        self.assertTrue(linux_out_dc1 in result.output)
        self.assertTrue(linux_out_dc2 in result.output)
        self.assertTrue(linux_out_dc3 in result.output)
        self.assertTrue('datacenter    : Bissen' in result.output)
        self.assertTrue('datacenter    : Level3 Baltimore' in result.output)
        self.assertTrue('datacenter    : Equinix Paris' in result.output)
        self.assertEqual(result.exit_code, 0)

    def test_kernels_match(self):

        result = self.invoke_with_exceptions(vm.kernels, ['3.10'])

        linux_out = """\
----------
flavor        : linux
version       : 3.10-x86_64
version       : 3.10-i386
"""
        self.assertTrue(linux_out in result.output)
        self.assertTrue('datacenter    : Bissen' in result.output)
        self.assertTrue('datacenter    : Level3 Baltimore' in result.output)
        self.assertTrue('datacenter    : Equinix Paris' in result.output)
        self.assertEqual(result.exit_code, 0)

    def test_kernels_flavor(self):
        args = ['--flavor', 'linux-hvm']
        result = self.invoke_with_exceptions(vm.kernels, args)

        self.assertEqual(result.output, """\
datacenter    : Equinix Paris
----------
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
version       : grub
version       : raw


datacenter    : Level3 Baltimore
----------
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
version       : grub
version       : raw


datacenter    : Bissen
----------
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
version       : grub
version       : raw


datacenter    : France, Paris
----------
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
version       : grub
version       : raw


datacenter    : France, Paris
----------
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
version       : grub
version       : raw
""")
        self.assertEqual(result.exit_code, 0)

    def test_kernels_datacenter(self):
        args = ['--datacenter', 'LU']
        result = self.invoke_with_exceptions(vm.kernels, args)

        linux_out = """\
flavor        : linux
version       : 2.6.32
version       : 2.6.27 (deprecated)
version       : 2.6.32-x86_64
version       : 2.6.32-x86_64-grsec
version       : 3.2-i386
version       : 3.2-x86_64
version       : 3.2-x86_64-grsec
version       : 3.10-x86_64
version       : 3.10-i386
"""
        linux_hvm_out = """\
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
version       : grub
version       : raw
"""
        self.assertTrue(linux_out in result.output)
        self.assertTrue(linux_hvm_out in result.output)
        self.assertTrue('datacenter    : Bissen' in result.output)
        self.assertEqual(result.exit_code, 0)

    def test_kernels_vm(self):

        result = self.invoke_with_exceptions(vm.kernels, ['--vm', 'server01'])

        linux_out = """\
flavor        : linux
version       : 2.6.18 (deprecated)
version       : 2.6.27-compat-sysfs (deprecated)
version       : 2.6.32
version       : 2.6.27 (deprecated)
version       : 2.6.32-x86_64
version       : 2.6.36 (deprecated)
version       : 2.6.32-x86_64-grsec
version       : 2.6.36-x86_64 (deprecated)
version       : 3.2-i386
version       : 3.2-x86_64
version       : 3.2-x86_64-grsec
version       : 3.10-x86_64
version       : 3.10-i386
"""
        linux_hvm_out = """\
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
version       : grub
version       : raw
"""
        self.assertTrue(linux_out in result.output)
        self.assertTrue(linux_hvm_out in result.output)
        self.assertTrue('datacenter    : Equinix Paris' in result.output)
        self.assertEqual(result.exit_code, 0)

    def test_kernels_all(self):

        args = ['--vm', 'server01', '--datacenter', 'FR',
                '--flavor', 'linux-hvm', '3.12']
        result = self.invoke_with_exceptions(vm.kernels, args)

        self.assertEqual(result.output, """\
datacenter    : Equinix Paris
----------
flavor        : linux-hvm
version       : 3.12-x86_64 (hvm)
""")
        self.assertEqual(result.exit_code, 0)

    def test_images(self):
        self.maxDiff = None
        result = self.invoke_with_exceptions(vm.images, [])

        self.assertEqual(result.output, """\
label         : Fedora 17 32 bits
os_arch       : x86-32
kernel_version: 3.2-i386
disk_id       : 527489
datacenter    : LU-BI1
----------
label         : Fedora 17 64 bits
os_arch       : x86-64
kernel_version: 3.2-x86_64
disk_id       : 527490
datacenter    : LU-BI1
----------
label         : OpenSUSE 12.2 32 bits
os_arch       : x86-32
kernel_version: 3.2-i386
disk_id       : 527491
datacenter    : LU-BI1
----------
label         : OpenSUSE 12.2 64 bits
os_arch       : x86-64
kernel_version: 3.2-x86_64
disk_id       : 527494
datacenter    : LU-BI1
----------
label         : CentOS 5 32 bits
os_arch       : x86-32
kernel_version: 2.6.32
disk_id       : 726224
datacenter    : LU-BI1
----------
label         : CentOS 5 64 bits
os_arch       : x86-64
kernel_version: 2.6.32-x86_64
disk_id       : 726225
datacenter    : LU-BI1
----------
label         : ArchLinux 32 bits
os_arch       : x86-32
kernel_version: 3.2-i386
disk_id       : 726230
datacenter    : LU-BI1
----------
label         : ArchLinux 64 bits
os_arch       : x86-64
kernel_version: 3.2-x86_64
disk_id       : 726233
datacenter    : LU-BI1
----------
label         : Debian 7 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 1401491
datacenter    : US-BA1
----------
label         : Debian 7 64 bits (HVM) /!\ DEPRECATED
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 1349810
datacenter    : FR-SD2
----------
label         : Debian 7 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 1401327
datacenter    : LU-BI1
----------
label         : Debian 8 (testing) 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3315704
datacenter    : FR-SD2
----------
label         : Debian 8 (testing) 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3315992
datacenter    : US-BA1
----------
label         : Debian 8
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3316070
datacenter    : FR-SD2
----------
label         : Debian 8
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3316070
datacenter    : LU-BI1
----------
label         : Debian 8
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3316070
datacenter    : FR-SD3
----------
label         : Debian 8
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3316070
datacenter    : FR-SD5
----------
label         : Debian 8 (testing) 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3316076
datacenter    : LU-BI1
----------
label         : Ubuntu 14.04 64 bits LTS (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3315748
datacenter    : FR-SD2
----------
label         : Ubuntu 14.04 64 bits LTS (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3316144
datacenter    : US-BA1
----------
label         : Ubuntu 14.04 64 bits LTS (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3316160
datacenter    : LU-BI1
----------
label         : CentOS 7 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 2876292
datacenter    : FR-SD2
----------
label         : CentOS 7 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 4744388
datacenter    : US-BA1
----------
label         : CentOS 7 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 4744392
datacenter    : LU-BI1
----------
label         : Debian 7 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 1401492
datacenter    : FR-SD3
----------
label         : Debian 7 64 bits (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 1401492
datacenter    : FR-SD5
----------
label         : Debian 7 64 bits (HVM)
kernel_version: 3.12-x86_64 (hvm)
name          : sys_1426759833
datacenter    : LU-BI1
----------
label         : Debian 7 64 bits (HVM)
kernel_version: 3.12-x86_64 (hvm)
name          : sys_server01
datacenter    : FR-SD2
----------
label         :
kernel_version:
name          : data
datacenter    : FR-SD2
----------
label         : Debian 7 64 bits
kernel_version: 3.2-x86_64
name          : snaptest
datacenter    : FR-SD2
----------
label         : Debian 7 64 bits (HVM)
kernel_version: 3.12-x86_64 (hvm)
name          : newdisk
datacenter    : LU-BI1
""")
        self.assertEqual(result.exit_code, 0)

    def test_images_all(self):

        args = ['Ubuntu 14.04', '--datacenter', 'LU']
        result = self.invoke_with_exceptions(vm.images, args)

        self.assertEqual(result.output, """\
label         : Ubuntu 14.04 64 bits LTS (HVM)
os_arch       : x86-64
kernel_version: 3.12-x86_64 (hvm)
disk_id       : 3316160
datacenter    : LU-BI1
""")
        self.assertEqual(result.exit_code, 0)

    def test_stop_one(self):
        result = self.invoke_with_exceptions(vm.stop, ['server01'])
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Stopping your Virtual Machine(s) 'server01'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_stop_multiple(self):
        args = ['server01', 'vm1426759833']
        result = self.invoke_with_exceptions(vm.stop, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Stopping your Virtual Machine(s) 'server01, vm1426759833'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_stop_background(self):
        result = self.invoke_with_exceptions(vm.stop, ['server01', '--bg'])
        self.assertEqual(result.output, """\
id        : 200
step      : WAIT
""")
        self.assertEqual(result.exit_code, 0)

    def test_start_one(self):
        result = self.invoke_with_exceptions(vm.start, ['server01'])
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Starting your Virtual Machine(s) 'server01'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_start_multiple(self):
        args = ['server01', 'vm1426759833']
        result = self.invoke_with_exceptions(vm.start, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Starting your Virtual Machine(s) 'server01, vm1426759833'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_start_background(self):
        result = self.invoke_with_exceptions(vm.start, ['server01', '--bg'])
        self.assertEqual(result.output, """\
id        : 200
step      : WAIT
""")
        self.assertEqual(result.exit_code, 0)

    def test_reboot_one(self):
        result = self.invoke_with_exceptions(vm.reboot, ['server01'])
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Rebooting your Virtual Machine(s) 'server01'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_reboot_multiple(self):
        args = ['server01', 'vm1426759833']
        result = self.invoke_with_exceptions(vm.reboot, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Rebooting your Virtual Machine(s) 'server01, vm1426759833'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_reboot_background(self):
        result = self.invoke_with_exceptions(vm.reboot, ['server01', '--bg'])
        self.assertEqual(result.output, """\
id        : 200
step      : WAIT
""")
        self.assertEqual(result.exit_code, 0)

    def test_delete_prompt(self):
        result = self.invoke_with_exceptions(vm.delete, ['server01'])
        self.assertEqual(result.output.strip(), """\
Are you sure to delete Virtual Machine 'server01'? [y/N]:""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_one(self):
        result = self.invoke_with_exceptions(vm.delete, ['server01', '-f'])
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Stopping your Virtual Machine(s) 'server01'.
\rProgress: [###] 100.00%  00:00:00  \n\
Deleting your Virtual Machine(s) 'server01'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_multiple(self):
        args = ['server01', 'vm1426759833', '-f']
        result = self.invoke_with_exceptions(vm.delete, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Stopping your Virtual Machine(s) 'server01'.
\rProgress: [###] 100.00%  00:00:00  \n\
Stopping your Virtual Machine(s) 'vm1426759833'.
\rProgress: [###] 100.00%  00:00:00  \n\
Deleting your Virtual Machine(s) 'server01, vm1426759833'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_unknown(self):
        result = self.invoke_with_exceptions(vm.delete, ['server100'])
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Sorry virtual machine server100 does not exist
Please use one of the following: ['vm1426759833', 'vm1426759844', 'server01', \
'server02', '152966', '152964', '152967', '152968']""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_background_ko(self):
        args = ['server01', '-f', '--bg']
        result = self.invoke_with_exceptions(vm.delete, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Virtual machine not stopped, background option disabled
Stopping your Virtual Machine(s) 'server01'.
\rProgress: [###] 100.00%  00:00:00  \n\
Deleting your Virtual Machine(s) 'server01'.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_background_ok(self):
        args = ['server02', '-f', '--bg']
        result = self.invoke_with_exceptions(vm.delete, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
id        : 200
step      : WAIT""")

        self.assertEqual(result.exit_code, 0)

    def test_update_ok(self):
        args = ['server01', '--memory', '1024', '--cores', '4']
        result = self.invoke_with_exceptions(vm.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your Virtual Machine server01.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_memory(self):
        args = ['server01', '--memory', '10240', '--cores', '4']
        result = self.invoke_with_exceptions(vm.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
memory update must be done offline.
reboot machine server01? [y/N]:""")

        self.assertEqual(result.exit_code, 0)

    def test_update_memory_reboot(self):
        args = ['server01', '--memory', '10240', '--cores', '4', '--reboot']
        result = self.invoke_with_exceptions(vm.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your Virtual Machine server01.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_background(self):
        args = ['server01', '--memory', '1024', '--cores', '4', '--bg']
        result = self.invoke_with_exceptions(vm.update, args)
        self.assertEqual(result.output, """\
{'id': 200, 'step': 'WAIT'}
""")
        self.assertEqual(result.exit_code, 0)

    def test_update_password(self):
        args = ['server01', '--password']
        result = self.invoke_with_exceptions(vm.update, args,
                                             input='plokiploki\nplokiploki\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \nUpdating your Virtual Machine server01.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_console(self):
        args = ['server01', '--console']
        result = self.invoke_with_exceptions(vm.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your Virtual Machine server01.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_console(self):
        args = ['server01']
        result = self.invoke_with_exceptions(vm.console, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\\ Please be aware that if you didn\'t provide a password during creation, \
console service will be unavailable.
/!\\ You can use "gandi vm update" command to set a password.
/!\\ Use ~. ssh escape key to exit.
Updating your Virtual Machine server01.
\rProgress: [###] 100.00%  00:00:00  \n\
ssh 95.142.160.181@console.gandi.net""")

        self.assertEqual(result.exit_code, 0)

    def test_ssh(self):
        args = ['admin@server01']
        result = self.invoke_with_exceptions(vm.ssh, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Requesting access using: ssh admin@95.142.160.181 ...
ssh admin@95.142.160.181""")

        self.assertEqual(result.exit_code, 0)

    def test_ssh_wipe_key(self):
        args = ['admin@server01', '--wipe-key']
        with mock.patch('gandi.cli.modules.iaas.open',
                        create=True) as mock_open:
            mock_open.return_value = mock.MagicMock()

            result = self.invoke_with_exceptions(vm.ssh, args)
            self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                    result.output.strip()), """\
Wiping old key and learning the new one
ssh-keygen -R "95.142.160.181"
Requesting access using: ssh admin@95.142.160.181 ...
ssh admin@95.142.160.181""")

            self.assertEqual(result.exit_code, 0)

    def test_ssh_wait(self):
        args = ['server01', '--wait']
        with mock.patch('gandi.cli.modules.iaas.socket',
                        create=True) as mock_socket:
            mock_socket.return_value = mock.MagicMock(name='socket',
                                                      spec=socket.socket)

            result = self.invoke_with_exceptions(vm.ssh, args)
            self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                    result.output.strip()), """\
Waiting for the vm to come online
Requesting access using: ssh root@95.142.160.181 ...
ssh root@95.142.160.181""")

            self.assertEqual(result.exit_code, 0)

    def test_ssh_login(self):
        args = ['server01', '--login', 'joe']
        result = self.invoke_with_exceptions(vm.ssh, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Requesting access using: ssh joe@95.142.160.181 ...
ssh joe@95.142.160.181""")

        self.assertEqual(result.exit_code, 0)

    def test_ssh_identity(self):
        args = ['admin@server01', '-i', 'key.pub']
        result = self.invoke_with_exceptions(vm.ssh, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Requesting access using: ssh -i key.pub admin@95.142.160.181 ...
ssh -i key.pub admin@95.142.160.181""")

        self.assertEqual(result.exit_code, 0)

    def test_ssh_args(self):
        args = ['server01', 'sudo reboot']
        result = self.invoke_with_exceptions(vm.ssh, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Requesting access using: ssh root@95.142.160.181 sudo reboot ...
ssh root@95.142.160.181 sudo reboot""")

        self.assertEqual(result.exit_code, 0)

    def test_create_default_hostname_ok(self):
        args = ['--hostname', 'server500']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \n* root user will be created.
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: server500, datacenter: FR-SD5
Creating your Virtual Machine server500.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine server500 has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_default_ok(self):
        args = []
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'vm\d+', 'vm', output), """\
password: \nRepeat for confirmation: \n* root user will be created.
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: vm, datacenter: FR-SD5
Creating your Virtual Machine vm.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine vm has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_ip_not_vlan_ko(self):
        args = ['--hostname', 'server500', '--ip', '10.50.10.10']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \n\
--ip can't be used without --vlan.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_vlan_ip_ok(self):
        args = ['--hostname', 'server400', '--vlan', 'vlantest',
                '--ip', '10.50.10.10']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \n* Private only ip vm (can't enable \
emergency web console access).
* root user will be created.
Creating your iface.
\rProgress: [###] 100.00%  00:00:00  \n\
Your iface has been created with the following IP addresses:
ip4:\t10.50.10.10
* Configuration used: 1 cores, 256Mb memory, ip private, image Debian 8\
, hostname: server400, datacenter: FR-SD5
Creating your Virtual Machine server400.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine server400 has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_login_ok(self):
        args = ['--login', 'administrator']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'vm\d+', 'vm', output), """\
password: \nRepeat for confirmation: \n\
* root and administrator users will be created.
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: vm, datacenter: FR-SD5
Creating your Virtual Machine vm.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine vm has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_background_ok(self):
        args = ['--hostname', 'server500', '--background']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \n* root user will be created.
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: server500, datacenter: FR-SD5
* IAAS backend is now creating your VM and its associated resources in the \
background.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_sshkey_ok(self):
        args = ['--sshkey', 'mysecretkey']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper())
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'vm\d+', 'vm', output), """\
* root user will be created.
* SSH key authorization will be used.
* No password supplied for vm (required to enable emergency web console \
access).
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: vm, datacenter: FR-SD5
Creating your Virtual Machine vm.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine vm has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_gen_password_root_ok(self):
        args = ['--gen-password']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper())
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        output = re.sub(r'vm\d+', 'vm', output)
        output = re.sub(r'with password .*', 'with password FAKEPASSWORD',
                        output)

        self.assertEqual(output, """\
* root user will be created.
* User root setup with password FAKEPASSWORD
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: vm, datacenter: FR-SD5
Creating your Virtual Machine vm.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine vm has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_gen_password_user_ok(self):
        args = ['--gen-password', '--login', 'myuser']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper())
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        output = re.sub(r'vm\d+', 'vm', output)
        output = re.sub(r'with password .*', 'with password FAKEPASSWORD',
                        output)

        self.assertEqual(output, """\
* root and myuser users will be created.
* Users root and myuser setup with password FAKEPASSWORD
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: vm, datacenter: FR-SD5
Creating your Virtual Machine vm.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine vm has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_image_deprecated(self):
        args = ['--image', 'Debian 7 64 bits (HVM)',
                '--sshkey', 'mysecretkey',
                '--datacenter', 'FR-SD2']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper())
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'vm\d+', 'vm', output), """\
/!\ Datacenter FR-SD2 will be closed on 25/12/2017, please consider \
using another datacenter.
/!\ Image Debian 7 64 bits (HVM) is deprecated and will soon be unavailable.
* root user will be created.
* SSH key authorization will be used.
* No password supplied for vm (required to enable emergency web console \
access).
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 7 64 bits \
(HVM), hostname: vm, datacenter: FR-SD2
Creating your Virtual Machine vm.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine vm has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_dc_code_ok(self):
        args = ['--datacenter', 'FR-SD3']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'vm\d+', 'vm', output), """\
password: \nRepeat for confirmation: \n* root user will be created.
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: vm, datacenter: FR-SD3
Creating your Virtual Machine vm.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine vm has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_datacenter_closed(self):
        args = ['--datacenter', 'US-BA1']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'vm\d+', 'vm', output), """\
Error: /!\ Datacenter US-BA1 is closed, please choose another datacenter.""")

        self.assertEqual(result.exit_code, 1)

    def test_create_datacenter_limited(self):
        args = ['--datacenter', 'FR-SD2']
        result = self.invoke_with_exceptions(vm.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')
        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'vm\d+', 'vm', output), """\
/!\ Datacenter FR-SD2 will be closed on 25/12/2017, please consider \
using another datacenter.
password: \nRepeat for confirmation: \n\
* root user will be created.
* Configuration used: 1 cores, 256Mb memory, ip v6, image Debian 8\
, hostname: vm, datacenter: FR-SD2
Creating your Virtual Machine vm.
\rProgress: [###] 100.00%  00:00:00  \n\
Your Virtual Machine vm has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_migrate_not_available(self):
        args = ['vm1426759844']
        result = self.invoke_with_exceptions(vm.migrate, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Your VM vm1426759844 cannot be migrated yet. \
Migration will be available when datacenter FR-SD5 is opened.""")

        self.assertEqual(result.exit_code, 0)

    def test_migrate_ok(self):
        args = ['server02', '-f']
        result = self.invoke_with_exceptions(vm.migrate, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
* Starting the migration of VM server02 from datacenter FR-SD2 to LU-BI1
VM migration in progress.
\rProgress: [###] 100.00%  00:00:00  \n\
Your VM server02 has been migrated.""")

        self.assertEqual(result.exit_code, 0)

    def test_migrate_noconfirm(self):
        args = ['server02']
        result = self.invoke_with_exceptions(vm.migrate, args, input='\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to migrate VM server02 ? [y/N]:""")

        self.assertEqual(result.exit_code, 0)

    def test_migrate_background(self):
        args = ['server02', '--bg', '-f']
        result = self.invoke_with_exceptions(vm.migrate, args)
        self.assertEqual(result.output.strip(), """\
* Starting the migration of VM server02 from datacenter FR-SD2 to LU-BI1
id        : 9900
step      : WAIT""")
        self.assertEqual(result.exit_code, 0)

    def test_migrate_finalize_not_found(self):
        args = ['server02', '-f', '--finalize']
        result = self.invoke_with_exceptions(vm.migrate, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Error: Cannot find VM server02 migration operation.""")

        self.assertEqual(result.exit_code, 1)

    def test_migrate_finalize_not_needed(self):
        args = ['server01', '-f', '--finalize']
        result = self.invoke_with_exceptions(vm.migrate, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Error: VM server01 migration does not need finalization.""")

        self.assertEqual(result.exit_code, 1)

    def test_migrate_finalize_ok(self):
        args = ['vm1426759833', '-f', '--finalize']
        result = self.invoke_with_exceptions(vm.migrate, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
* Finalizing the migration of VM vm1426759833 from datacenter FR-SD2 to LU-BI1
VM migration in progress.
\rProgress: [###] 100.00%  00:00:00  \n\
Your VM vm1426759833 has been migrated.""")

        self.assertEqual(result.exit_code, 0)



""" Hosting account module. """

from gandi.cli.core.base import GandiModule


class Account(GandiModule):

    """ Module to handle CLI commands.

    $ gandi account info

    """

    @classmethod
    def info(cls):
        """Get information about the hosting account in use"""
        return cls.call('hosting.account.info')

    @classmethod
    def creditusage(cls):
        """Get credit usage per hour"""
        rating = cls.call('hosting.rating.list')
        if not rating:
            return 0

        rating = rating.pop()
        usage = [sum(resource.values())
                 for resource in rating.values()
                 if isinstance(resource, dict)]
        return sum(usage)

    @classmethod
    def all(cls):
        """ Get all informations about this account """
        account = cls.info()
        creditusage = cls.creditusage()

        if not creditusage:
            return account

        left = account['credits'] / creditusage
        years, hours = divmod(left, 365 * 24)
        months, hours = divmod(hours, 31 * 24)
        days, hours = divmod(hours, 24)

        account.update({'credit_usage': creditusage,
                        'left': (years, months, days, hours)})

        return account



""" Vhost commands module. """

import os

from gandi.cli.core.base import GandiModule


class Vhost(GandiModule):

    """ Module to handle CLI commands.

    $ gandi vhost create
    $ gandi vhost delete
    $ gandi vhost info
    $ gandi vhost list

    """

    @classmethod
    def list(cls, options=None):
        """ List paas vhosts (in the future it should handle iaas vhosts)."""
        options = options or {}
        return cls.call('paas.vhost.list', options)

    @classmethod
    def info(cls, name):
        """ Display information about a vhost. """
        return cls.call('paas.vhost.info', name)

    @classmethod
    def create(cls, paas_info, vhost, alter_zone, background):
        """ Create a new vhost. """
        if not background and not cls.intty():
            background = True

        params = {'paas_id': paas_info['id'],
                  'vhost': vhost,
                  'zone_alter': alter_zone}
        result = cls.call('paas.vhost.create', params, dry_run=True)

        if background:
            return result

        cls.echo('Creating a new vhost.')
        cls.display_progress(result)
        cls.echo('Your vhost %s has been created.' % vhost)

        return result

    @classmethod
    def delete(cls, resources, background=False):
        """ Delete this vhost. """
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        opers = []
        for item in resources:
            oper = cls.call('paas.vhost.delete', item)
            opers.append(oper)

        if background:
            return opers

        cls.echo('Deleting your vhost.')
        cls.display_progress(opers)



""" Operation commands module. """

from gandi.cli.core.base import GandiModule


class Oper(GandiModule):

    """ Module to handle CLI commands.

    $ gandi oper info
    $ gandi oper list

    """

    @classmethod
    def list(cls, options):
        """List operation."""
        return cls.call('operation.list', options)

    @classmethod
    def count(cls, options):
        """Count operation."""
        return cls.call('operation.count', options)

    @classmethod
    def info(cls, id):
        """Display information about an operation."""
        return cls.call('operation.info', id)



""" Mail commands module. """

from gandi.cli.core.base import GandiModule


class Mail(GandiModule):

    """ Module to handle CLI commands.

    $ gandi mail create
    $ gandi mail delete
    $ gandi mail info
    $ gandi mail list
    $ gandi mail purge
    $ gandi mail update

    """

    @classmethod
    def list(cls, domain, options):
        """List mailboxes for a given domain name."""
        return cls.call('domain.mailbox.list', domain, options)

    @classmethod
    def info(cls, domain, login):
        """Display information about a mailbox."""
        return cls.call('domain.mailbox.info', domain, login)

    @classmethod
    def create(cls, domain, login, options, alias):
        """Create a mailbox."""
        cls.echo('Creating your mailbox.')
        result = cls.call('domain.mailbox.create', domain, login, options)

        if alias:
            cls.echo('Creating aliases.')
            result = cls.set_alias(domain, login, list(alias))

        return result

    @classmethod
    def delete(cls, domain, login):
        """Delete a mailbox."""
        return cls.call('domain.mailbox.delete', domain, login)

    @classmethod
    def update(cls, domain, login, options, alias_add, alias_del):
        """Update a mailbox."""
        cls.echo('Updating your mailbox.')
        result = cls.call('domain.mailbox.update', domain, login, options)

        if alias_add or alias_del:
            current_aliases = cls.info(domain, login)['aliases']
            aliases = current_aliases[:]
            if alias_add:
                for alias in alias_add:
                    if alias not in aliases:
                        aliases.append(alias)
            if alias_del:
                for alias in alias_del:
                    if alias in aliases:
                        aliases.remove(alias)

            if ((len(current_aliases) != len(aliases))
                    or (current_aliases != aliases)):
                cls.echo('Updating aliases.')
                result = cls.set_alias(domain, login, aliases)

        return result

    @classmethod
    def purge(cls, domain, login, background=False):
        """Purge a mailbox."""
        oper = cls.call('domain.mailbox.purge', domain, login)
        if background:
            return oper
        else:
            cls.echo('Purging in progress')
            cls.display_progress(oper)

    @classmethod
    def set_alias(cls, domain, login, aliases):
        """Update aliases on a mailbox."""
        return cls.call('domain.mailbox.alias.set', domain, login, aliases)



from .base import CommandTestCase
from gandi.cli.commands import sshkey


class SSHKeyTestCase(CommandTestCase):

    def test_list(self):
        args = ['--id']
        result = self.invoke_with_exceptions(sshkey.list, args)

        self.assertEqual(result.output, """\
name        : default
fingerprint : b3:11:67:10:2e:1b:a5:66:ed:16:24:98:3e:2e:ed:f5
id          : 134
----------
name        : mysecretkey
fingerprint : 09:11:21:e3:90:3c:7d:d5:06:d9:6f:f9:36:e1:99:a6
id          : 141
""")
        self.assertEqual(result.exit_code, 0)

    def test_info(self):
        args = ['default', '--id', '--value']
        result = self.invoke_with_exceptions(sshkey.info, args)

        self.assertEqual(result.output, """\
name        : default
fingerprint : b3:11:67:10:2e:1b:a5:66:ed:16:24:98:3e:2e:ed:f5
id          : 134
value       : ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC63QZAW3tusdv+JuyzOoXTND9/wxKogMwZbxBPPtoN7Hjnyn0kUUHMJ6ji5xpbatRYKOeGAoZDW2TXojvbJdQj7tWsRr7ES0qB9qhDGVSDIJWRQ6f9MQCCLjV5tpBTAwb unknown@lol.cat
""")  # noqa
        self.assertEqual(result.exit_code, 0)

    def test_delete(self):
        args = ['mysecretkey']
        result = self.invoke_with_exceptions(sshkey.delete, args)

        self.assertEqual(result.output, '')
        self.assertEqual(result.exit_code, 0)

    def test_create_value_and_filename_ko(self):
        args = ['--name', 'newkey', '--value',
                'ssh-rsa LjV5tpBTAwb unknown@inter.net',
                '--filename', 'sandbox/example.txt']

        content = """\
ssh-rsa LjV5tpBTAwb unknown@inter.net
"""
        result = self.isolated_invoke_with_exceptions(sshkey.create, args,
                                                      temp_content=content)

        self.assertEqual(result.output, """\
Usage: create [OPTIONS]

Error: You must not set value AND filename.
""")
        self.assertEqual(result.exit_code, 2)

    def test_create_no_value_and_no_filename_ko(self):
        args = ['--name', 'newkey']

        result = self.invoke_with_exceptions(sshkey.create, args)

        self.assertEqual(result.output, """\
Usage: create [OPTIONS]

Error: You must set value OR filename.
""")
        self.assertEqual(result.exit_code, 2)

    def test_create_value_ok(self):
        args = ['--name', 'newkey', '--value',
                'ssh-rsa LjV5tpBTAwb unknown@inter.net']

        result = self.invoke_with_exceptions(sshkey.create, args)

        self.assertEqual(result.output, """\
id          : 145
name        : newkey
fingerprint : b3:11:67:10:2e:1b:a5:55:ed:16:24:98:3e:2e:ed:f5
""")
        self.assertEqual(result.exit_code, 0)

    def test_create_file_ok(self):
        args = ['--name', 'newkey', '--filename', 'sandbox/example.txt']

        content = """\
ssh-rsa LjV5tpBTAwb unknown@inter.net
"""
        result = self.isolated_invoke_with_exceptions(sshkey.create, args,
                                                      temp_content=content)

        self.assertEqual(result.output, """\
id          : 145
name        : newkey
fingerprint : b3:11:67:10:2e:1b:a5:55:ed:16:24:98:3e:2e:ed:f5
""")
        self.assertEqual(result.exit_code, 0)



""" Disk commands module. """

from gandi.cli.core.base import GandiModule
from gandi.cli.core.utils import DuplicateResults
from gandi.cli.core.params import DISK_MAXLIST
from .iaas import Iaas, Datacenter, Image


class Disk(GandiModule):

    """ Module to handle CLI commands.

    $ gandi disk create
    $ gandi disk delete
    $ gandi disk attach
    $ gandi disk detach
    $ gandi disk info
    $ gandi disk list
    $ gandi disk rollback
    $ gandi disk snapshot
    $ gandi disk update

    """

    @classmethod
    def from_name(cls, name):
        """ Retrieve a disk id associated to a name. """
        disks = cls.list({'name': name})
        if len(disks) == 1:
            return disks[0]['id']
        elif not disks:
            return

        raise DuplicateResults('disk name %s is ambiguous.' % name)

    @classmethod
    def usable_id(cls, id):
        """ Retrieve id from input which can be name or id."""
        try:
            qry_id = cls.from_name(id)
            if not qry_id:
                qry_id = int(id)
        except DuplicateResults as exc:
            cls.error(exc.errors)
        except Exception:
            qry_id = None

        if not qry_id:
            msg = 'unknown identifier %s' % id
            cls.error(msg)

        return qry_id

    @classmethod
    def list(cls, options=None):
        """ List all disks."""
        options = options or {}
        return cls.call('hosting.disk.list', options)

    @classmethod
    def list_create(cls, datacenter=None, label=None):
        """List available disks for vm creation."""
        options = {
            'items_per_page': DISK_MAXLIST
        }
        if datacenter:
            datacenter_id = int(Datacenter.usable_id(datacenter))
            options['datacenter_id'] = datacenter_id

        # implement a filter by label as API doesn't handle it
        images = cls.safe_call('hosting.disk.list', options)
        if not label:
            return images
        return [img for img in images
                if label.lower() in img['name'].lower()]

    @classmethod
    def _info(cls, disk_id):
        """ Get information about a disk."""
        return cls.call('hosting.disk.info', disk_id)

    @classmethod
    def info(cls, name):
        """ Get information about a disk."""
        return cls._info(cls.usable_id(name))

    @staticmethod
    def disk_param(name, size, snapshot_profile, cmdline=None, kernel=None):
        """ Return disk parameter structure. """
        disk_params = {}

        if cmdline:
            disk_params['cmdline'] = cmdline

        if kernel:
            disk_params['kernel'] = kernel

        if name:
            disk_params['name'] = name

        if snapshot_profile is not None:
            disk_params['snapshot_profile'] = snapshot_profile

        if size:
            disk_params['size'] = size

        return disk_params

    @classmethod
    def update(cls, resource, name, size, snapshot_profile,
               background, cmdline=None, kernel=None):
        """ Update this disk. """
        if isinstance(size, tuple):
            prefix, size = size
            if prefix == '+':
                disk_info = cls.info(resource)
                current_size = disk_info['size']
                size = current_size + size

        disk_params = cls.disk_param(name, size, snapshot_profile,
                                     cmdline, kernel)

        result = cls.call('hosting.disk.update',
                          cls.usable_id(resource),
                          disk_params)
        if background:
            return result

        # interactive mode, run a progress bar
        cls.echo('Updating your disk.')
        cls.display_progress(result)

    @classmethod
    def _detach(cls, disk_id):
        """ Detach a disk from a vm. """
        disk = cls._info(disk_id)
        opers = []
        if disk.get('vms_id'):
            for vm_id in disk['vms_id']:
                cls.echo('The disk is still attached to the vm %s.' % vm_id)
                cls.echo('Will detach it.')
                opers.append(cls.call('hosting.vm.disk_detach',
                                      vm_id, disk_id))
        return opers

    @classmethod
    def detach(cls, resources, background):
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        resources = [cls.usable_id(item) for item in resources]

        opers = []
        for disk_id in resources:
            opers.extend(cls._detach(disk_id))

        if opers and not background:
            cls.echo('Detaching your disk(s).')
            cls.display_progress(opers)

        return opers

    @classmethod
    def delete(cls, resources, background=False):
        """ Delete this disk."""
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        resources = [cls.usable_id(item) for item in resources]

        opers = []
        for disk_id in resources:
            opers.extend(cls._detach(disk_id))

        if opers:
            cls.echo('Detaching your disk(s).')
            cls.display_progress(opers)

        opers = []
        for disk_id in resources:
            oper = cls.call('hosting.disk.delete', disk_id)
            opers.append(oper)

        if background:
            return opers

        cls.echo('Deleting your disk.')
        cls.display_progress(opers)

        return opers

    @classmethod
    def _attach(cls, disk_id, vm_id, options=None):
        """ Attach a disk to a vm. """
        options = options or {}
        oper = cls.call('hosting.vm.disk_attach', vm_id, disk_id, options)
        return oper

    @classmethod
    def attach(cls, disk, vm, background, position=None, read_only=False):
        from gandi.cli.modules.iaas import Iaas as VM
        vm_id = VM.usable_id(vm)

        disk_id = cls.usable_id(disk)
        disk_info = cls._info(disk_id)

        options = {}
        if position is not None:
            options['position'] = position

        if read_only:
            options['access'] = 'read'

        need_detach = disk_info.get('vms_id')
        if need_detach:
            if disk_info.get('vms_id') == [vm_id]:
                cls.echo('This disk is already attached to this vm.')
                return

            # detach disk
            detach_op = cls._detach(disk_id)

            # interactive mode, run a progress bar
            cls.echo('Detaching your disk.')
            cls.display_progress(detach_op)

        oper = cls._attach(disk_id, vm_id, options)

        if oper and not background:
            cls.echo('Attaching your disk(s).')
            cls.display_progress(oper)

        return oper

    @classmethod
    def create(cls, name, vm, size, snapshotprofile, datacenter,
               source, disk_type='data', background=False):
        """ Create a disk and attach it to a vm. """
        if isinstance(size, tuple):
            prefix, size = size

        if source:
            size = None
        disk_params = cls.disk_param(name, size, snapshotprofile)
        disk_params['datacenter_id'] = int(Datacenter.usable_id(datacenter))
        disk_params['type'] = disk_type

        if source:
            disk_id = int(Image.usable_id(source,
                                          disk_params['datacenter_id']))
            result = cls.call('hosting.disk.create_from', disk_params, disk_id)
        else:
            result = cls.call('hosting.disk.create', disk_params)

        if background and not vm:
            return result

        # interactive mode, run a progress bar
        cls.echo('Creating your disk.')
        cls.display_progress(result)

        if not vm:
            return

        vm_id = Iaas.usable_id(vm)
        result = cls._attach(result['disk_id'], vm_id)

        if background:
            return result

        cls.echo('Attaching your disk.')
        cls.display_progress(result)

    @classmethod
    def rollback(cls, resource, background=False):
        """ Rollback a disk from a snapshot. """
        disk_id = cls.usable_id(resource)
        result = cls.call('hosting.disk.rollback_from', disk_id)

        if background:
            return result

        cls.echo('Disk rollback in progress.')
        cls.display_progress(result)
        return result

    @classmethod
    def migrate(cls, resource, datacenter_id, background=False):
        """ Migrate a disk to another datacenter. """
        disk_id = cls.usable_id(resource)
        result = cls.call('hosting.disk.migrate', disk_id, datacenter_id)

        if background:
            return result

        cls.echo('Disk migration in progress.')
        cls.display_progress(result)
        return result



# -*- coding: utf-8 -*-
""" XML-RPC connection helper. """

from __future__ import print_function

import sys
import socket

try:
    import xmlrpclib
except ImportError:
    import xmlrpc.client as xmlrpclib

try:
    import requests
except ImportError:
    print('python requests is required, please reinstall.', file=sys.stderr)
    sys.exit(1)

from gandi.cli import __version__
from gandi.cli.core.utils.xmlrpc import RequestsTransport


class APICallFailed(Exception):

    """ Raise when an error occurred during an API call. """

    def __init__(self, errors, code=None):
        """ Initialize exception. """
        self.errors = errors
        self.code = code


class GandiTransport(RequestsTransport):

    """ Mixin to send custom User-Agent in requests."""

    user_agent = 'gandi.cli/%s' % __version__


class DryRunException(APICallFailed):
    dry_run = None

    def __init__(self, message, code, dry_run):
        super(DryRunException, self).__init__(message, code)
        self.dry_run = dry_run


class XMLRPCClient(object):

    """ Class wrapper for xmlrpc calls to Gandi public API. """

    def __init__(self, host, debug=False):
        """ Initialize xml-rpc endpoint connector. """
        self.debug = debug
        self.host = host
        self.endpoint = xmlrpclib.ServerProxy(
            host, allow_none=True, use_datetime=True,
            transport=GandiTransport(use_datetime=True, host=host))

    def request(self, method, apikey, *args, **kwargs):
        """ Make a xml-rpc call to remote API. """
        dry_run = kwargs.get('dry_run', False)
        return_dry_run = kwargs.get('return_dry_run', False)
        if return_dry_run:
            args[-1]['--dry-run'] = True

        try:
            func = getattr(self.endpoint, method)
            return func(apikey, *args)
        except (socket.error, requests.exceptions.ConnectionError):
            msg = 'Gandi API service is unreachable'
            raise APICallFailed(msg)
        except xmlrpclib.Fault as err:
            msg = 'Gandi API has returned an error: %s' % err
            if dry_run:
                args[-1]['--dry-run'] = True
                ret = func(apikey, *args)
                raise DryRunException(msg, err.faultCode, ret)
            raise APICallFailed(msg, err.faultCode)
        except TypeError as err:
            msg = 'An unknown error has occurred: %s' % err
            raise APICallFailed(msg)


class JsonClient(object):

    """ Class wrapper for JSON calls. """

    @classmethod
    def format_errors(cls, errors):
        return '\n'.join([err['description'] for err in errors])

    @classmethod
    def request(cls, method, url, **kwargs):
        """Make a http call to a remote API and return a json response."""
        user_agent = 'gandi.cli/%s' % __version__

        headers = {'User-Agent': user_agent,
                   'Content-Type': 'application/json; charset=utf-8'}
        if kwargs.get('headers'):
            headers.update(kwargs.pop('headers'))
        try:
            response = requests.request(method, url, headers=headers,
                                        **kwargs)
            response.raise_for_status()
            try:
                return response.json(), response.headers
            except ValueError as err:
                return response.text, response.headers
        except (socket.error, requests.exceptions.ConnectionError):
            msg = 'Remote API service is unreachable'
            raise APICallFailed(msg)
        except Exception as err:
            if isinstance(err, requests.HTTPError):
                try:
                    resp = response.json()
                except Exception:
                    msg = 'An unknown error has occurred: %s' % err
                    raise APICallFailed(msg)

                if resp.get('message'):
                    error = resp.get('message')
                if resp.get('errors'):
                    error = cls.format_errors(resp.get('errors'))
                msg = '%s: %s' % (err, error)
            else:
                msg = 'An unknown error has occurred: %s' % err
            raise APICallFailed(msg)



# -*- coding: utf-8 -*-
import re
import sys
from datetime import datetime

from .base import CommandTestCase
from ..compat import mock
from gandi.cli.commands import domain
from gandi.cli.core.utils import DomainNotAvailable


class DomainTestCase(CommandTestCase):

    def test_list(self):

        result = self.invoke_with_exceptions(domain.list, [])

        self.assertEqual(result.output, """iheartcli.com
cli.sexy
""")
        self.assertEqual(result.exit_code, 0)

    def test_info(self):
        with mock.patch('gandi.cli.core.utils.datetime') as mock_datetime:
            mock_datetime.now.return_value = datetime(2015, 7, 1)
            mock_datetime.side_effect = lambda *args, **kw: datetime(*args,
                                                                     **kw)

            result = self.invoke_with_exceptions(domain.info,
                                                 ['iheartcli.com'])

        self.assertEqual(result.output, """owner       : AA1-GANDI
admin       : AA2-GANDI
bill        : AA3-GANDI
tech        : AA5-GANDI
reseller    : AA4-GANDI
fqdn        : iheartcli.com
nameservers : a.dns.gandi.net, b.dns.gandi.net, c.dns.gandi.net
services    : gandidns
zone_id     : 424242
tags        : bla
created     : 2010-09-22 15:06:18
expires     : 2015-09-22 00:00:00 (in 83 days)
updated     : 2014-09-21 03:10:07
""")
        self.assertEqual(result.exit_code, 0)

    def test_create_no_option_no_argument(self):
        args = ['--duration', 1,
                '--owner', 'OWNER1-GANDI',
                '--admin', 'ADMIN1-GANDI',
                '--tech', 'TECH1-GANDI',
                '--bill', 'BILL1-GANDI',
                ]
        result = self.invoke_with_exceptions(domain.create, args,
                                             input='idontlike.website\n')

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
Name of the domain: idontlike.website
Creating your domain.
\rProgress: [###] 100.00%  00:00:00  \n\
Your domain idontlike.website has been created.""")

        self.assertEqual(result.exit_code, 0)
#         self.assertEqual(result.output.strip(), """\
# Name of the domain.:\
# """)

    def test_create_option(self):
        result = self.invoke_with_exceptions(domain.create,
                                             ['--domain', 'idontlike.website',
                                              '--duration', 1,
                                              '--owner', 'OWNER1-GANDI',
                                              '--admin', 'ADMIN1-GANDI',
                                              '--tech', 'TECH1-GANDI',
                                              '--bill', 'BILL1-GANDI',
                                              '--nameserver', 'a.domain.tld',
                                              '--nameserver', 'b.domain.tld',
                                              ])

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
/!\ --domain option is deprecated and will be removed upon next release.
You should use 'gandi domain create idontlike.website' instead.
Creating your domain.
\rProgress: [###] 100.00%  00:00:00  \n\
Your domain idontlike.website has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_argument(self):
        result = self.invoke_with_exceptions(domain.create,
                                             ['idontlike.website',
                                              '--duration', 1,
                                              '--owner', 'OWNER1-GANDI',
                                              '--admin', 'ADMIN1-GANDI',
                                              '--tech', 'TECH1-GANDI',
                                              '--bill', 'BILL1-GANDI',
                                              ])

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
Creating your domain.
\rProgress: [###] 100.00%  00:00:00  \n\
Your domain idontlike.website has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_argument_and_option_different(self):
        result = self.invoke_with_exceptions(domain.create,
                                             ['idontlike.website',
                                              '--domain', 'idontlike.bike',
                                              '--duration', 1,
                                              '--owner', 'OWNER1-GANDI',
                                              '--admin', 'ADMIN1-GANDI',
                                              '--tech', 'TECH1-GANDI',
                                              '--bill', 'BILL1-GANDI',
                                              ])

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
/!\ --domain option is deprecated and will be removed upon next release.
You should use 'gandi domain create idontlike.bike' instead.
/!\ You specified both an option and an argument which are different, \
please choose only one between: idontlike.bike and idontlike.website.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_argument_and_option_equal(self):
        result = self.invoke_with_exceptions(domain.create,
                                             ['idontlike.website',
                                              '--domain', 'idontlike.website',
                                              '--duration', 1,
                                              '--owner', 'OWNER1-GANDI',
                                              '--admin', 'ADMIN1-GANDI',
                                              '--tech', 'TECH1-GANDI',
                                              '--bill', 'BILL1-GANDI',
                                              ])

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
/!\ --domain option is deprecated and will be removed upon next release.
You should use 'gandi domain create idontlike.website' instead.
Creating your domain.
\rProgress: [###] 100.00%  00:00:00  \n\
Your domain idontlike.website has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_background_argument(self):
        args = ['roflozor.com', '--background']
        result = self.invoke_with_exceptions(domain.create, args)

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
Duration [1]: \n\
{'id': 400, 'step': 'WAIT'}""")

        self.assertEqual(result.exit_code, 0)

    def test_create_background_option(self):
        args = ['--domain', 'roflozor.com', '--background']
        result = self.invoke_with_exceptions(domain.create, args)

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
Duration [1]: \n\
/!\ --domain option is deprecated and will be removed upon next release.
You should use 'gandi domain create roflozor.com' instead.
{'id': 400, 'step': 'WAIT'}""")

        self.assertEqual(result.exit_code, 0)

    def test_renew(self):
        result = self.invoke_with_exceptions(domain.renew,
                                             ['iheartcli.com',
                                              '--duration', 1,
                                              ])

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
Renewing your domain.
\rProgress: [###] 100.00%  00:00:00  \n\
Your domain iheartcli.com has been renewed.""")

        self.assertEqual(result.exit_code, 0)

    def test_renew_background_ok(self):
        args = ['iheartcli.com', '--background']
        result = self.invoke_with_exceptions(domain.renew, args)

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        self.assertEqual(output, """\
Duration [1]: \n\
{'id': 400, 'step': 'WAIT'}""")

        self.assertEqual(result.exit_code, 0)

    def test_available_with_exception(self):

        self.assertRaises(DomainNotAvailable,
                          self.invoke_with_exceptions, domain.create,
                          ['--domain', 'unavailable1.website',
                           '--duration', 1,
                           '--owner', 'OWNER1-GANDI',
                           '--admin', 'ADMIN1-GANDI',
                           '--tech', 'TECH1-GANDI',
                           '--bill', 'BILL1-GANDI',
                           ])

    def test_available_with_exception_argument(self):
        self.assertRaises(DomainNotAvailable,
                          self.invoke_with_exceptions, domain.create,
                          ['unavailable1.website',
                           '--duration', 1,
                           '--owner', 'OWNER1-GANDI',
                           '--admin', 'ADMIN1-GANDI',
                           '--tech', 'TECH1-GANDI',
                           '--bill', 'BILL1-GANDI',
                           ])

    def test_create_background_argument_extra_parameter(self):
        args = ['roflozor.com', '--background',
                '--extra_parameter', 'x-aero_ens_authid', '1234',
                '--extra_parameter', 'x-aero_ens_authkey', 'password']
        result = self.invoke_with_exceptions(domain.create, args)

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())
        if sys.version_info[0] == 2:
            self.assertEqual(output, """\
Duration [1]: \n{'extra': {u'x-aero_ens_authid': u'1234', u'x-aero_ens_authkey': u'password'},\n 'id': 400,\n 'step': 'WAIT'}""")  # noqa
        else:
            self.assertEqual(output, """\
Duration [1]: \n{'extra': {'x-aero_ens_authid': '1234', 'x-aero_ens_authkey': 'password'},\n 'id': 400,\n 'step': 'WAIT'}""")  # noqa

        self.assertEqual(result.exit_code, 0)



# -*- coding: utf-8 -*-
import re

from .base import CommandTestCase
from gandi.cli.commands import ip
from gandi.cli.core.base import GandiContextHelper


class IpTestCase(CommandTestCase):

    def test_list(self):

        result = self.invoke_with_exceptions(ip.list, [])

        self.assertEqual(result.output, """\
ip         : 95.142.160.181
state      : created
type       : public
datacenter : FR-SD2
----------
ip         : 2001:4b98:dc2:43:216:3eff:fece:e25f
state      : created
type       : public
datacenter : LU-BI1
----------
ip         : 2001:4b98:dc0:47:216:3eff:feb2:3862
state      : created
type       : public
datacenter : FR-SD2
----------
ip         : 192.168.232.253
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
----------
ip         : 192.168.232.252
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_details(self):
        args = ['--id', '--version', '--vm', '--reverse']
        result = self.invoke_with_exceptions(ip.list, args)

        self.assertEqual(result.output, """\
ip         : 95.142.160.181
state      : created
id         : 203968
version    : 4
reverse    : xvm-160-181.dc0.ghst.net
type       : public
vm         : server01
datacenter : FR-SD2
----------
ip         : 2001:4b98:dc2:43:216:3eff:fece:e25f
state      : created
id         : 204557
version    : 6
reverse    : xvm6-dc2-fece-e25f.ghst.net
type       : public
datacenter : LU-BI1
----------
ip         : 2001:4b98:dc0:47:216:3eff:feb2:3862
state      : created
id         : 204558
version    : 6
reverse    : xvm6-dc0-feb2-3862.ghst.net
type       : public
vm         : server01
datacenter : FR-SD2
----------
ip         : 192.168.232.253
state      : created
id         : 2361
version    : 4
reverse    :
type       : private
vlan       : pouet
vm         : server02
datacenter : FR-SD2
----------
ip         : 192.168.232.252
state      : created
id         : 2361
version    : 4
reverse    :
type       : private
vlan       : pouet
vm         : server02
datacenter : FR-SD2
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_attached(self):

        result = self.invoke_with_exceptions(ip.list, ['--attached'])

        self.assertEqual(result.output, """\
ip         : 95.142.160.181
state      : created
type       : public
datacenter : FR-SD2
----------
ip         : 2001:4b98:dc0:47:216:3eff:feb2:3862
state      : created
type       : public
datacenter : FR-SD2
----------
ip         : 192.168.232.253
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
----------
ip         : 192.168.232.252
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_detached(self):

        result = self.invoke_with_exceptions(ip.list, ['--detached'])

        self.assertEqual(result.output, """\
ip         : 2001:4b98:dc2:43:216:3eff:fece:e25f
state      : created
type       : public
datacenter : LU-BI1
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_filter_type(self):

        result = self.invoke_with_exceptions(ip.list, ['--type', 'private'])

        self.assertEqual(result.output, """\
ip         : 2001:4b98:dc2:43:216:3eff:fece:e25f
state      : created
type       : public
datacenter : LU-BI1
----------
ip         : 192.168.232.253
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
----------
ip         : 192.168.232.252
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_filter_datacenter(self):

        result = self.invoke_with_exceptions(ip.list, ['--datacenter', 'FR'])

        self.assertEqual(result.output, """\
ip         : 95.142.160.181
state      : created
type       : public
datacenter : FR-SD2
----------
ip         : 2001:4b98:dc0:47:216:3eff:feb2:3862
state      : created
type       : public
datacenter : FR-SD2
----------
ip         : 192.168.232.253
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
----------
ip         : 192.168.232.252
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_filter_vlan(self):

        result = self.invoke_with_exceptions(ip.list, ['--vlan', 'pouet'])

        self.assertEqual(result.output, """\
ip         : 192.168.232.253
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
----------
ip         : 192.168.232.252
state      : created
type       : private
vlan       : pouet
datacenter : FR-SD2
""")
        self.assertEqual(result.exit_code, 0)

    def test_list_attached_detached(self):
        args = ['--detached', '--attached']
        result = self.invoke_with_exceptions(ip.list, args)

        self.assertEqual(result.output, """\
You can't set --attached and --detached at the same time.
""")
        self.assertEqual(result.exit_code, 0)

    def test_info(self):
        args = ['95.142.160.181']
        result = self.invoke_with_exceptions(ip.info, args)

        self.assertEqual(result.output, """\
ip         : 95.142.160.181
state      : created
reverse    : xvm-160-181.dc0.ghst.net
type       : public
vm         : server01
datacenter : FR-SD2
""")
        self.assertEqual(result.exit_code, 0)

    def test_update_ko(self):
        args = ['95.142.160.181']
        result = self.invoke_with_exceptions(ip.update, args)

        self.assertEqual(result.output, '')
        self.assertEqual(result.exit_code, 0)

    def test_update_reverse(self):
        args = ['95.142.160.181', '--reverse', 'plop.bloup.com']
        result = self.invoke_with_exceptions(ip.update, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your IP
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_attach_ko(self):
        args = ['395.142.160.181', 'vm1426759833']
        result = self.invoke_with_exceptions(ip.attach, args)

        self.assertTrue("Can't find this ip 395.142.160.181" in result.output)
        self.assertEqual(result.exit_code, 2)

    def test_attach_already(self):
        args = ['95.142.160.181', 'server01']
        result = self.invoke_with_exceptions(ip.attach, args, input='y\n')

        self.assertEqual(result.output, """\
This ip is already attached to this vm.
""")
        self.assertEqual(result.exit_code, 0)

    def test_attach(self):
        args = ['95.142.160.181', 'vm1426759833']
        result = self.invoke_with_exceptions(ip.attach, args, input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to detach 95.142.160.181 from vm 152967 [y/N]: y
The iface is still attached to the vm 152967.
Will detach it.
\rProgress: [###] 100.00%  00:00:00  \
\n\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_attach_force(self):
        args = ['95.142.160.181', 'vm1426759833', '--force']
        result = self.invoke_with_exceptions(ip.attach, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
The iface is still attached to the vm 152967.
Will detach it.
\rProgress: [###] 100.00%  00:00:00  \
\n\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_attach_refuse(self):
        args = ['95.142.160.181', 'vm1426759833']
        result = self.invoke_with_exceptions(ip.attach, args, input='N\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to detach 95.142.160.181 from vm 152967 [y/N]: N""")

        self.assertEqual(result.exit_code, 0)

    def test_attach_background(self):
        args = ['95.142.160.181', 'vm1426759833', '--force', '--bg']
        result = self.invoke_with_exceptions(ip.attach, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
The iface is still attached to the vm 152967.
Will detach it.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_create_default(self):
        args = []
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Creating your iface.
\rProgress: [###] 100.00%  00:00:00  \
\nYour iface has been created with the following IP addresses:
ip4:\t95.142.160.181
ip6:\t2001:4b98:dc0:47:216:3eff:feb2:3862""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.iface.create'][0][0]
        self.assertEqual(params['datacenter_id'], 3)
        self.assertEqual(params['bandwidth'], 102400)
        self.assertEqual(params['ip_version'], 4)

    def test_create_datacenter_limited(self):
        args = ['--datacenter', 'FR-SD2', '--bandwidth', '51200',
                '--ip-version', '6']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ Datacenter FR-SD2 will be closed on 25/12/2017, please consider using \
another datacenter.
Creating your iface.
\rProgress: [###] 100.00%  00:00:00  \
\nYour iface has been created with the following IP addresses:
ip4:\t95.142.160.181
ip6:\t2001:4b98:dc0:47:216:3eff:feb2:3862""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.iface.create'][0][0]
        self.assertEqual(params['datacenter_id'], 1)
        self.assertEqual(params['bandwidth'], 51200)
        self.assertEqual(params['ip_version'], 6)

    def test_create_datacenter_closed(self):
        args = ['--datacenter', 'US-BA1', '--bandwidth', '51200',
                '--ip-version', '6']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Error: /!\ Datacenter US-BA1 is closed, please choose another datacenter.""")

        self.assertEqual(result.exit_code, 1)

    def test_create_params(self):
        args = ['--datacenter', 'FR', '--bandwidth', '51200',
                '--ip-version', '6']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ Datacenter FR will be closed on 25/12/2017, please consider using \
another datacenter.
Creating your iface.
\rProgress: [###] 100.00%  00:00:00  \
\nYour iface has been created with the following IP addresses:
ip4:\t95.142.160.181
ip6:\t2001:4b98:dc0:47:216:3eff:feb2:3862""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.iface.create'][0][0]
        self.assertEqual(params['datacenter_id'], 1)
        self.assertEqual(params['bandwidth'], 51200)
        self.assertEqual(params['ip_version'], 6)

    def test_create_params_vlan_ko(self):
        args = ['--datacenter', 'FR', '--bandwidth', '51200',
                '--ip-version', '6', '--vlan', 'pouet']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
You must have an --ip-version to 4 when having a vlan.""")
        self.assertEqual(result.exit_code, 0)

    def test_create_params_vlan_ok(self):
        args = ['--datacenter', 'FR', '--bandwidth', '51200',
                '--ip-version', '4', '--vlan', 'pouet']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ Datacenter FR will be closed on 25/12/2017, please consider using \
another datacenter.
Creating your iface.
\rProgress: [###] 100.00%  00:00:00  \
\nYour iface has been created with the following IP addresses:
ip4:\t95.142.160.181
ip6:\t2001:4b98:dc0:47:216:3eff:feb2:3862""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.iface.create'][0][0]
        self.assertEqual(params['datacenter_id'], 1)
        self.assertEqual(params['bandwidth'], 51200)
        self.assertEqual(params['ip_version'], 4)
        self.assertEqual(params['vlan'], 717)

    def test_create_params_ip_ko(self):
        args = ['--datacenter', 'FR', '--bandwidth', '51200',
                '--ip-version', '4', '--ip', '10.50.10.10']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
You must have a --vlan when giving an --ip.""")
        self.assertEqual(result.exit_code, 0)

    def test_create_params_ip_ok(self):
        args = ['--datacenter', 'FR', '--bandwidth', '51200',
                '--ip-version', '4', '--ip', '10.50.10.10',
                '--vlan', 'pouet']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ Datacenter FR will be closed on 25/12/2017, please consider using \
another datacenter.
Creating your iface.
\rProgress: [###] 100.00%  00:00:00  \
\nYour iface has been created with the following IP addresses:
ip4:\t10.50.10.10""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.iface.create'][0][0]
        self.assertEqual(params['datacenter_id'], 1)
        self.assertEqual(params['bandwidth'], 51200)
        self.assertEqual(params['ip_version'], 4)
        self.assertEqual(params['vlan'], 717)
        self.assertEqual(params['ip'], '10.50.10.10')

    def test_create_params_attach_ko(self):
        args = ['--datacenter', 'US', '--bandwidth', '51200',
                '--ip-version', '4', '--attach', 'server01']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
The datacenter you provided does not match the datacenter of the \
vm you want to attach to.""")
        self.assertEqual(result.exit_code, 0)

    def test_create_params_attach_ok(self):
        args = ['--datacenter', 'FR', '--bandwidth', '51200',
                '--ip-version', '4', '--ip', '10.50.10.10',
                '--vlan', 'pouet', '--attach', 'server01']
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ Datacenter FR will be closed on 25/12/2017, please consider using \
another datacenter.
Creating your iface.
\rProgress: [###] 100.00%  00:00:00  \
\nYour iface has been created with the following IP addresses:
ip4:\t10.50.10.10
Attaching your iface.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.iface.create'][0][0]
        self.assertEqual(params['datacenter_id'], 1)
        self.assertEqual(params['bandwidth'], 51200)
        self.assertEqual(params['ip_version'], 4)
        self.assertEqual(params['vlan'], 717)
        self.assertEqual(params['ip'], '10.50.10.10')

    def test_create_background(self):
        args = ['--datacenter', 'FR', '--bandwidth', '51200',
                '--ip-version', '4', '--ip', '10.50.10.10',
                '--vlan', 'pouet', '--attach', 'server01',
                '--background']
        self.maxDiff = None
        result = self.invoke_with_exceptions(ip.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ Datacenter FR will be closed on 25/12/2017, please consider using \
another datacenter.
Creating your iface.
\rProgress: [###] 100.00%  00:00:00  \
\nYour iface has been created with the following IP addresses:
ip4:\t10.50.10.10""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.iface.create'][0][0]
        self.assertEqual(params['datacenter_id'], 1)
        self.assertEqual(params['bandwidth'], 51200)
        self.assertEqual(params['ip_version'], 4)
        self.assertEqual(params['vlan'], 717)
        self.assertEqual(params['ip'], '10.50.10.10')

    def test_detach(self):
        args = ['95.142.160.181']
        result = self.invoke_with_exceptions(ip.detach, args, input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to detach ip 95.142.160.181? [y/N]: y\
\nThe iface is still attached to the vm 152967.
Will detach it.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_detach_refuse(self):
        args = ['95.142.160.181']
        result = self.invoke_with_exceptions(ip.detach, args, input='N\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to detach ip 95.142.160.181? [y/N]: N""")

        self.assertEqual(result.exit_code, 0)

    def test_detach_force(self):
        args = ['95.142.160.181', '--force']
        result = self.invoke_with_exceptions(ip.detach, args, input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
The iface is still attached to the vm 152967.
Will detach it.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_unknown(self):
        args = ['395.142.160.181']
        result = self.invoke_with_exceptions(ip.delete, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Sorry interface 395.142.160.181 does not exist
Please use one of the following: ['203968', '204557', '204558', '2361', \
'2361', '95.142.160.181', '2001:4b98:dc2:43:216:3eff:fece:e25f', \
'2001:4b98:dc0:47:216:3eff:feb2:3862', '192.168.232.253', '192.168.232.252'\
]""")

        self.assertEqual(result.exit_code, 0)

    def test_delete(self):
        args = ['95.142.160.181']
        result = self.invoke_with_exceptions(ip.delete, args, input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to delete ip(s) 95.142.160.181 [y/N]: y
The iface is still attached to the vm 152967.
Will detach it.
Detaching your iface(s).
\rProgress: [###] 100.00%  00:00:00  \
\nDetaching/deleting your iface(s).
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_refuse(self):
        args = ['95.142.160.181']
        result = self.invoke_with_exceptions(ip.delete, args, input='N\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to delete ip(s) 95.142.160.181 [y/N]: N""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_force(self):
        args = ['95.142.160.181', '--force']
        result = self.invoke_with_exceptions(ip.delete, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
The iface is still attached to the vm 152967.
Will detach it.
Detaching your iface(s).
\rProgress: [###] 100.00%  00:00:00  \
\nDetaching/deleting your iface(s).
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_background(self):
        args = ['95.142.160.181', '--background']
        result = self.invoke_with_exceptions(ip.delete, args, input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to delete ip(s) 95.142.160.181 [y/N]: y
The iface is still attached to the vm 152967.
Will detach it.
Detaching your iface(s).
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_multiple(self):
        args = ['95.142.160.181', '2001:4b98:dc2:43:216:3eff:fece:e25f']
        result = self.invoke_with_exceptions(ip.delete, args, input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to delete ip(s) 2001:4b98:dc2:43:216:3eff:fece:e25f, \
95.142.160.181 [y/N]: y
The iface is still attached to the vm 152967.
Will detach it.
Detaching your iface(s).
\rProgress: [###] 100.00%  00:00:00  \
\nDetaching/deleting your iface(s).
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)



from datetime import datetime

try:
    # python3
    from xmlrpc.client import DateTime
except ImportError:
    # python2
    from xmlrpclib import DateTime

type_list = list


def list(options):
    return [{'authinfo': 'abcdef0001',
             'autorenew': None,
             'zone_id': 424242,
             'tags': 'bla',
             'contacts': {'owner': {'handle': 'AA1-GANDI'},
                          'admin': {'handle': 'AA2-GANDI'},
                          'bill': {'handle': 'AA3-GANDI'},
                          'reseller': {'handle': 'AA4-GANDI'},
                          'tech': {'handle': 'AA5-GANDI'}},
             'date_created': datetime(2010, 9, 22, 15, 6, 18),
             'date_delete': datetime(2015, 10, 19, 19, 14, 0),
             'date_hold_begin': datetime(2015, 9, 22, 22, 0, 0),
             'date_registry_creation': datetime(2010, 9, 22, 13, 6, 16),
             'date_registry_end': datetime(2015, 9, 22, 0, 0, 0),
             'date_updated': datetime(2014, 9, 21, 3, 10, 7),
             'nameservers': ['a.dns.gandi.net', 'b.dns.gandi.net',
                             'c.dns.gandi.net'],
             'services': ['gandidns'],
             'fqdn': 'iheartcli.com',
             'id': 236816922,
             'status': [],
             'tld': 'com'},
            {'authinfo': 'abcdef0002',
             'autorenew': None,
             'contacts': {'admin': {'handle': 'PXP561-GANDI', 'id': 2920674},
                          'bill': {'handle': 'PXP561-GANDI', 'id': 2920674},
                          'owner': {'handle': 'PXP561-GANDI', 'id': 2920674},
                          'reseller': None,
                          'tech': {'handle': 'PXP561-GANDI', 'id': 2920674}},
             'date_created': DateTime('20130410T12:46:05'),
             'date_delete': DateTime('20160507T07:14:00'),
             'date_hold_begin': DateTime('20160410T00:00:00'),
             'date_registry_creation': DateTime('20140410T10:46:04'),
             'date_registry_end': DateTime('20140410T00:00:00'),
             'date_updated': DateTime('20150313T10:30:05'),
             'date_hold_end': DateTime('20151020T20:00:00'),
             'date_pending_delete_end': DateTime('20151119T00:00:00'),
             'date_renew_begin': DateTime('20120101T00:00:00'),
             'date_restore_end': DateTime('20151119T00:00:00'),
             'fqdn': 'cli.sexy',
             'id': 3412062241,
             'nameservers': ['a.dns.gandi.net', 'b.dns.gandi.net',
                             'c.dns.gandi.net'],
             'services': ['gandidns', 'gandimail', 'paas'],
             'status': [],
             'tags': [],
             'tld': 'sexy',
             'zone_id': None}]


def info(id):
    domain = dict([(domain['fqdn'], domain) for domain in list({})])
    return domain[id]


def available(domains):

    ret = {}
    for domain in domains:
        if 'unavailable' in domain:
            ret[domain] = 'unavailable'
        elif 'pending' in domain:
            ret[domain] = 'pending'
        else:
            ret[domain] = 'available'

    return ret


def create(domain, params):
    ret = {'id': 400, 'step': 'WAIT'}

    if params and 'extra' in params:
        ret['extra'] = params['extra']

    return ret


def renew(domain, params):
    return {'id': 400, 'step': 'WAIT'}


def mailbox_list(domain, options):
    return [{'login': 'admin',
             'responder': {'active': False},
             'quota': {'granted': 0, 'used': 233}}]


def mailbox_info(domain, login):
    ret = {'aliases': [],
           'login': 'admin',
           'responder': {'active': False, 'text': None},
           'fallback_email': '',
           'quota': {'granted': 0, 'used': 233}}
    return ret


def mailbox_create(domain, login, params):
    return {'id': 400, 'step': 'WAIT'}


def mailbox_delete(domain, login):
    return {'id': 400, 'step': 'WAIT'}


def mailbox_update(domain, login, params):
    return {'id': 400, 'step': 'WAIT'}


def mailbox_alias_set(domain, login, aliases):
    return {'id': 400, 'step': 'WAIT'}


def mailbox_purge(domain, login):
    return {'id': 400, 'step': 'WAIT'}


def forward_list(domain, options):
    return [{'source': 'admin',
             'destinations': ['admin@cli.sexy', 'grumpy@cat.lol']},
            {'source': 'contact',
             'destinations': ['contact@cli.sexy']}]


def forward_create(domain, source, options):
    return [{'source': source,
             'destinations': options['destinations']}]


def forward_update(domain, source, options):
    return [{'source': source,
             'destinations': options['destinations']}]


def forward_delete(domain, source):
    return True


def zone_record_list(zone_id, version, options=None):
    ret = [{'id': 337085079,
            'name': '*',
            'ttl': 10800,
            'type': 'A',
            'value': '73.246.104.110'},
           {'id': 337085078,
            'name': '@',
            'ttl': 10800,
            'type': 'A',
            'value': '73.246.104.110'},
           {'id': 337085081,
            'name': 'much',
            'ttl': 10800,
            'type': 'A',
            'value': '192.243.24.132'},
           {'id': 337085072,
            'name': 'blog',
            'ttl': 10800,
            'type': 'CNAME',
            'value': 'blogs.vip.gandi.net.'},
           {'id': 337085082,
            'name': 'cloud',
            'ttl': 10800,
            'type': 'CNAME',
            'value': 'gpaas6.dc0.gandi.net.'},
           {'id': 337085075,
            'name': 'imap',
            'ttl': 10800,
            'type': 'CNAME',
            'value': 'access.mail.gandi.net.'},
           {'id': 337085071,
            'name': 'pop',
            'ttl': 10800,
            'type': 'CNAME',
            'value': 'access.mail.gandi.net.'},
           {'id': 337085074,
            'name': 'smtp',
            'ttl': 10800,
            'type': 'CNAME',
            'value': 'relay.mail.gandi.net.'},
           {'id': 337085073,
            'name': 'webmail',
            'ttl': 10800,
            'type': 'CNAME',
            'value': 'agent.mail.gandi.net.'},
           {'id': 337085077,
            'name': '@',
            'ttl': 10800,
            'type': 'MX',
            'value': '50 fb.mail.gandi.net.'},
           {'id': 337085076,
            'name': '@',
            'ttl': 10800,
            'type': 'MX',
            'value': '10 spool.mail.gandi.net.'}]

    options = options or {}
    options.pop('items_per_page', None)

    def match(zone, options):
        for fkey in options:
            if zone[fkey] != options[fkey]:
                return

        return zone

    ret = [zone for zone in ret if match(zone, options)]

    return ret


def zone_version_new(zone_id):
    return 242424


def zone_record_add(zone_id, version, data):
    return


def zone_version_set(zone_id, version):
    return


def zone_record_delete(zone_id, version, data):
    return


def zone_record_update(zone_id, version, opts, data):
    return


def zone_record_set(zone_id, version, data):
    return


def dnssec_delete(key_id):
    return {'id': 666, 'step': 'WAIT'}


def dnssec_list(domain):
    return [{'algorithm': 5,
             'date_created': datetime(2012, 2, 24, 17, 16, 8),
             'digest': '457c626c008cc70d68133254abc4ee4eb79e4e6c99f9423b60b543'
                       'fa8a69e6ac',
             'digest_type': 2,
             'flags': 257,
             'id': 125,
             'keytag': 9301,
             'public_key': 'AwEAAdYixYvq9eJLRQcxUeYJWaxAGXiP/K1/C7XHbUWGzA8AHC'
                           'Rp81FAmfwcw1FrJ7bMViEegewPDGciQSv5HotPPOynUmkZbgzt'
                           'OeejH/+3Il/cM8SW4Et0i+99S7l9as+FI3AYOhsllDJK1WM9sm'
                           'n0S/9igfpR2dGmCyDU ZfeR1A49\n'}]


def dnssec_create(domain, params):
    return {'id': 667, 'step': 'WAIT'}



# -*- coding: utf-8 -*-

import re

from .base import CommandTestCase
from gandi.cli.commands import vhost


class VhostTestCase(CommandTestCase):

    def test_list(self):
        args = ['--id', '--names']
        result = self.invoke_with_exceptions(vhost.list, args)

        self.assertEqual(result.output, """\
name          : aa3e0e26f8.url-de-test.ws
state         : running
date_creation : 20130903T22:11:54
paas_id       : 126276
paas_name     : paas_owncloud
----------
name          : cloud.cat.lol
state         : running
date_creation : 20130903T22:24:06
paas_id       : 126276
paas_name     : paas_owncloud
----------
name          : 187832c2b34.testurl.ws
state         : running
date_creation : 20141025T15:50:54
paas_id       : 163744
paas_name     : paas_cozycloud
----------
name          : cloud.iheartcli.com
state         : running
date_creation : 20141025T15:50:54
paas_id       : 163744
paas_name     : paas_cozycloud
----------
name          : cli.sexy
state         : running
date_creation : 20150728T17:50:56
paas_id       : 163744
paas_name     : paas_cozycloud
""")
        self.assertEqual(result.exit_code, 0)

    def test_info(self):
        args = ['cloud.cat.lol', 'cloud.iheartcli.com', '--id']
        result = self.invoke_with_exceptions(vhost.info, args)

        self.assertEqual(result.output, """\
name          : cloud.cat.lol
state         : running
date_creation : 20130903T22:24:06
ssl           : disabled
paas_id       : 126276
paas_name     : paas_owncloud
----------
name          : cloud.iheartcli.com
state         : running
date_creation : 20141025T15:50:54
ssl           : disabled
paas_id       : 163744
paas_name     : paas_cozycloud
""")
        self.assertEqual(result.exit_code, 0)

    def test_create(self):
        args = ['pouet.lol.cat', '--paas', 'paas_owncloud']
        result = self.invoke_with_exceptions(vhost.create, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Creating a new vhost.
\rProgress: [###] 100.00%  00:00:00  \
\nYour vhost pouet.lol.cat has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_ssl(self):
        args = ['pouet.lol.cat', '--paas', 'paas_owncloud', '--ssl']
        result = self.invoke_with_exceptions(vhost.create, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Please give the private key for certificate id 710 (CN: lol.cat)""")

        self.assertEqual(result.exit_code, 0)

    def test_create_background(self):
        args = ['pouet.lol.cat', '--paas', 'paas_owncloud', '--bg']
        result = self.invoke_with_exceptions(vhost.create, args)

        self.assertEqual(result.output.strip(), """\
{'id': 200, 'step': 'WAIT'}""")

        self.assertEqual(result.exit_code, 0)

    def test_update_ssl_ko(self):
        args = ['pouet.lol.cat', '--ssl']
        result = self.invoke_with_exceptions(vhost.update, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Please give the private key for certificate id 710 (CN: lol.cat)""")

        self.assertEqual(result.exit_code, 0)

    def test_update_ssl_ok(self):
        args = ['unknown.lol.cat', '--ssl']
        result = self.invoke_with_exceptions(vhost.update, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
There is no certificate for unknown.lol.cat.
Create the certificate with (for exemple) :
$ gandi certificate create --cn unknown.lol.cat --type std \
\nOr relaunch the current command with --poll-cert option""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_prompt_ok(self):
        args = ['pouet.lol.cat']
        result = self.invoke_with_exceptions(vhost.delete, args,
                                             input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to delete vhost pouet.lol.cat? [y/N]: y
Deleting your vhost.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_prompt_ko(self):
        args = ['pouet.lol.cat']
        result = self.invoke_with_exceptions(vhost.delete, args,
                                             input='N\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to delete vhost pouet.lol.cat? [y/N]: N""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_force(self):
        args = ['pouet.lol.cat', '--force']
        result = self.invoke_with_exceptions(vhost.delete, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Deleting your vhost.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_background(self):
        args = ['pouet.lol.cat', '--force', '--bg']
        result = self.invoke_with_exceptions(vhost.delete, args)

        self.assertEqual(result.output.strip(), """\
name      : rproxy_update
paas_id   : 1177220
date_creation: 20150728T17:50:56""")
        self.assertEqual(result.exit_code, 0)



""" VM commands module. """

import math
import os
import socket
import time
import errno

from gandi.cli.core.base import GandiModule
from gandi.cli.core.utils import randomstring
from gandi.cli.modules.datacenter import Datacenter
from gandi.cli.modules.sshkey import SshkeyHelper
from gandi.cli.core.utils import MigrationNotFinalized


class Iaas(GandiModule, SshkeyHelper):

    """ Module to handle CLI commands.

    $ gandi vm console
    $ gandi vm create
    $ gandi vm delete
    $ gandi vm images
    $ gandi vm info
    $ gandi vm kernels
    $ gandi vm list
    $ gandi vm reboot
    $ gandi vm ssh
    $ gandi vm start
    $ gandi vm stop
    $ gandi vm update

    """

    @classmethod
    def list(cls, options=None):
        """List virtual machines."""
        if not options:
            options = {}

        return cls.call('hosting.vm.list', options)

    @classmethod
    def resource_list(cls):
        """ Get the possible list of resources (hostname, id). """
        items = cls.list({'items_per_page': 500})
        ret = [vm['hostname'] for vm in items]
        ret.extend([str(vm['id']) for vm in items])
        return ret

    @classmethod
    def info(cls, id):
        """Display information about a virtual machine."""
        return cls.call('hosting.vm.info', cls.usable_id(id))

    @classmethod
    def stop(cls, resources, background=False):
        """Stop a virtual machine."""
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        opers = []
        for item in resources:
            oper = cls.call('hosting.vm.stop', cls.usable_id(item))
            if isinstance(oper, list):
                opers.extend(oper)
            else:
                opers.append(oper)

        if background:
            return opers

        # interactive mode, run a progress bar
        instance_info = "'%s'" % ', '.join(resources)
        cls.echo('Stopping your Virtual Machine(s) %s.' % instance_info)
        cls.display_progress(opers)

    @classmethod
    def start(cls, resources, background=False):
        """Start a virtual machine."""
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        opers = []
        for item in resources:
            oper = cls.call('hosting.vm.start', cls.usable_id(item))
            if isinstance(oper, list):
                opers.extend(oper)
            else:
                opers.append(oper)

        if background:
            return opers

        # interactive mode, run a progress bar
        instance_info = "'%s'" % ', '.join(resources)
        cls.echo('Starting your Virtual Machine(s) %s.' % instance_info)
        cls.display_progress(opers)

    @classmethod
    def reboot(cls, resources, background=False):
        """Reboot a virtual machine."""
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        opers = []
        for item in resources:
            oper = cls.call('hosting.vm.reboot', cls.usable_id(item))
            if isinstance(oper, list):
                opers.extend(oper)
            else:
                opers.append(oper)

        if background:
            return opers

        # interactive mode, run a progress bar
        instance_info = "'%s'" % ', '.join(resources)
        cls.echo('Rebooting your Virtual Machine(s) %s.' % instance_info)
        cls.display_progress(opers)

    @classmethod
    def delete(cls, resources, background=False):
        """Delete a virtual machine."""
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        opers = []
        for item in resources:
            oper = cls.call('hosting.vm.delete', cls.usable_id(item))
            if not oper:
                continue

            if isinstance(oper, list):
                opers.extend(oper)
            else:
                opers.append(oper)

        if background:
            return opers

        # interactive mode, run a progress bar
        instance_info = "'%s'" % ', '.join(resources)
        cls.echo('Deleting your Virtual Machine(s) %s.' % instance_info)
        if opers:
            cls.display_progress(opers)

    @classmethod
    def required_max_memory(cls, id, memory):
        """
        Recommend a max_memory setting for this vm given memory. If the
        VM already has a nice setting, return None. The max_memory
        param cannot be fixed too high, because page table allocation
        would cost too much for small memory profile. Use a range as below.
        """
        best = int(max(2 ** math.ceil(math.log(memory, 2)), 2048))

        actual_vm = cls.info(id)

        if (actual_vm['state'] == 'running'
                and actual_vm['vm_max_memory'] != best):
            return best

    @classmethod
    def update(cls, id, memory, cores, console, password, background,
               max_memory):
        """Update a virtual machine."""
        if not background and not cls.intty():
            background = True

        vm_params = {}

        if memory:
            vm_params['memory'] = memory

        if cores:
            vm_params['cores'] = cores

        if console:
            vm_params['console'] = console

        if password:
            vm_params['password'] = password

        if max_memory:
            vm_params['vm_max_memory'] = max_memory

        result = cls.call('hosting.vm.update', cls.usable_id(id), vm_params)
        if background:
            return result

        # interactive mode, run a progress bar
        cls.echo('Updating your Virtual Machine %s.' % id)
        cls.display_progress(result)

    @classmethod
    def create(cls, datacenter, memory, cores, ip_version, bandwidth,
               login, password, hostname, image, run, background, sshkey,
               size, vlan, ip, script, script_args, ssh):
        """Create a new virtual machine."""
        from gandi.cli.modules.network import Ip, Iface
        if not background and not cls.intty():
            background = True

        datacenter_id_ = int(Datacenter.usable_id(datacenter))

        if not hostname:
            hostname = randomstring('vm')
            disk_name = 'sys_%s' % hostname[2:]
        else:
            disk_name = 'sys_%s' % hostname.replace('.', '')

        vm_params = {
            'hostname': hostname,
            'datacenter_id': datacenter_id_,
            'memory': memory,
            'cores': cores,
        }

        if login:
            vm_params['login'] = login

        if run:
            vm_params['run'] = run

        if password:
            vm_params['password'] = password

        if ip_version:
            vm_params['ip_version'] = ip_version
            vm_params['bandwidth'] = bandwidth

        if script:
            with open(script) as fd:
                vm_params['script'] = fd.read()
            if script_args:
                vm_params['script_args'] = script_args

        vm_params.update(cls.convert_sshkey(sshkey))

        # XXX: name of disk is limited to 15 chars in ext2fs, ext3fs
        # but api allow 255, so we limit to 15 for now
        disk_params = {'datacenter_id': vm_params['datacenter_id'],
                       'name': disk_name[:15]}

        if size:
            if isinstance(size, tuple):
                prefix, size = size
            disk_params['size'] = size

        sys_disk_id_ = int(Image.usable_id(image, datacenter_id_))

        ip_summary = []
        if ip_version == 4:
            ip_summary = ['v4', 'v6']
        elif ip_version == 6:
            ip_summary = ['v6']

        if vlan:
            ip_ = None
            ip_summary.append('private')
            if ip:
                try:
                    ip_ = Ip.info(ip)
                except Exception:
                    pass
                else:
                    if not Ip._check_and_detach(ip_, None):
                        return
            if ip_:
                iface_id = ip_['iface_id']
            else:
                ip_create = Ip.create(4,
                                      vm_params['datacenter_id'],
                                      bandwidth,
                                      None,
                                      vlan,
                                      ip)

                iface_id = ip_create['iface_id']

            # if there is a public ip, will attach this one later, else give
            # the iface to vm.create
            if not ip_version:
                vm_params['iface_id'] = iface_id

        result = cls.call('hosting.vm.create_from', vm_params, disk_params,
                          sys_disk_id_)

        cls.echo('* Configuration used: %d cores, %dMb memory, ip %s, '
                 'image %s, hostname: %s, datacenter: %s' %
                 (cores, memory, '+'.join(ip_summary), image, hostname,
                  datacenter))

        # background mode, bail out now (skip interactive part)
        if background and (not vlan or not ip_version):
            return result

        # interactive mode, run a progress bar
        cls.echo('Creating your Virtual Machine %s.' % hostname)
        cls.display_progress(result)
        cls.echo('Your Virtual Machine %s has been created.' % hostname)

        vm_id = None
        for oper in result:
            if oper.get('vm_id'):
                vm_id = oper.get('vm_id')
                break

        if vlan and ip_version:
            attach = Iface._attach(iface_id, vm_id)
            if background:
                return attach

        if 'ssh_key' not in vm_params and 'keys' not in vm_params:
            return

        if vm_id and ip_version:
            cls.wait_for_sshd(vm_id)
            if ssh:
                cls.ssh_keyscan(vm_id)
                cls.ssh(vm_id, 'root', None)

    @classmethod
    def need_finalize(cls, resource):
        """Check if vm migration need to be finalized."""
        vm_id = cls.usable_id(resource)
        params = {'type': 'hosting_migration_vm',
                  'step': 'RUN',
                  'vm_id': vm_id}
        result = cls.call('operation.list', params)
        if not result or len(result) > 1:
            raise MigrationNotFinalized('Cannot find VM %s '
                                        'migration operation.' % resource)

        need_finalize = result[0]['params']['inner_step'] == 'wait_finalize'
        if not need_finalize:
            raise MigrationNotFinalized('VM %s migration does not need '
                                        'finalization.' % resource)

    @classmethod
    def check_can_migrate(cls, resource):
        """Check if virtual machine can be migrated to another datacenter."""
        vm_id = cls.usable_id(resource)
        result = cls.call('hosting.vm.can_migrate', vm_id)

        if not result['can_migrate']:
            if result['matched']:
                matched = result['matched'][0]
                cls.echo('Your VM %s cannot be migrated yet. Migration will '
                         'be available when datacenter %s is opened.'
                         % (resource, matched))
            else:
                cls.echo('Your VM %s cannot be migrated.' % resource)
            return False

        return True

    @classmethod
    def migrate(cls, resource, background=False, finalize=False):
        """ Migrate a virtual machine to another datacenter. """
        vm_id = cls.usable_id(resource)
        if finalize:
            verb = 'Finalizing'
            result = cls.call('hosting.vm.migrate', vm_id, True)
        else:
            verb = 'Starting'
            result = cls.call('hosting.vm.migrate', vm_id, False)

        dcs = {}
        for dc in Datacenter.list():
            dcs[dc['id']] = dc['dc_code']

        oper = cls.call('operation.info', result['id'])
        dc_from = dcs[oper['params']['from_dc_id']]
        dc_to = dcs[oper['params']['to_dc_id']]
        migration_msg = ('* %s the migration of VM %s '
                         'from datacenter %s to %s'
                         % (verb, resource, dc_from, dc_to))
        cls.echo(migration_msg)

        if background:
            return result

        cls.echo('VM migration in progress.')
        cls.display_progress(result)
        cls.echo('Your VM %s has been migrated.' % resource)
        return result

    @classmethod
    def from_hostname(cls, hostname):
        """Retrieve virtual machine id associated to a hostname."""
        result = cls.list({'hostname': str(hostname)})
        if result:
            return result[0]['id']

    @classmethod
    def usable_id(cls, id):
        """ Retrieve id from input which can be hostname or id."""
        try:
            # id is maybe a hostname
            qry_id = cls.from_hostname(id)
            if not qry_id:
                qry_id = int(id)
        except Exception:
            qry_id = None

        if not qry_id:
            msg = 'unknown identifier %s' % id
            cls.error(msg)

        return qry_id

    @classmethod
    def vm_ip(cls, vm_id):
        """Return the first usable ip address for this vm.
        Returns a (version, ip) tuple."""
        vm_info = cls.info(vm_id)

        for iface in vm_info['ifaces']:
            if iface['type'] == 'private':
                continue
            for ip in iface['ips']:
                return ip['version'], ip['ip']

    @classmethod
    def wait_for_sshd(cls, vm_id):
        """Insist on having the vm booted and sshd
        listening"""
        cls.echo('Waiting for the vm to come online')
        version, ip_addr = cls.vm_ip(vm_id)
        give_up = time.time() + 300
        last_error = None
        while time.time() < give_up:
            try:
                inet = socket.AF_INET
                if version == 6:
                    inet = socket.AF_INET6
                sd = socket.socket(inet, socket.SOCK_STREAM,
                                   socket.IPPROTO_TCP)
                sd.settimeout(5)
                sd.connect((ip_addr, 22))
                sd.recv(1024)
                return
            except socket.error as err:
                if err.errno == errno.EHOSTUNREACH and version == 6:
                    cls.error('%s is not reachable, you may be missing '
                              'IPv6 connectivity' % ip_addr)
                last_error = err
                time.sleep(1)
            except Exception as err:
                last_error = err
                time.sleep(1)
        cls.error('VM did not spin up (last error: %s)' % last_error)

    @classmethod
    def ssh_keyscan(cls, vm_id):
        """Wipe this old key and learn the new one from a freshly
        created vm. This is a security risk for this VM, however
        we dont have another way to learn the key yet, so do this
        for the user."""
        cls.echo('Wiping old key and learning the new one')
        _version, ip_addr = cls.vm_ip(vm_id)
        cls.execute('ssh-keygen -R "%s"' % ip_addr)

        for _ in range(5):
            output = cls.exec_output('ssh-keyscan "%s"' % ip_addr)
            if output:
                with open(os.path.expanduser('~/.ssh/known_hosts'), 'a') as f:
                    f.write(output)
                return True
            time.sleep(.5)

    @classmethod
    def scp(cls, vm_id, login, identity, local_file, remote_file):
        """Copy file to remote VM."""
        cmd = ['scp']
        if identity:
            cmd.extend(('-i', identity,))

        version, ip_addr = cls.vm_ip(vm_id)
        if version == 6:
            ip_addr = '[%s]' % ip_addr

        cmd.extend((local_file, '%s@%s:%s' %
                    (login, ip_addr, remote_file),))
        cls.echo('Running %s' % ' '.join(cmd))
        for _ in range(5):
            ret = cls.execute(cmd, False)
            if ret:
                break
            time.sleep(.5)
        return ret

    @classmethod
    def ssh(cls, vm_id, login, identity, args=None):
        """Spawn an ssh session to virtual machine."""
        cmd = ['ssh']
        if identity:
            cmd.extend(('-i', identity,))

        version, ip_addr = cls.vm_ip(vm_id)
        if version == 6:
            cmd.append('-6')

        if not ip_addr:
            cls.echo('No IP address found for vm %s, aborting.' % vm_id)
            return

        cmd.append('%s@%s' % (login, ip_addr,))

        if args:
            cmd.extend(args)

        cls.echo('Requesting access using: %s ...' % ' '.join(cmd))
        return cls.execute(cmd, False)

    @classmethod
    def console(cls, id):
        """Open a console to virtual machine."""
        vm_info = cls.info(id)
        if not vm_info['console']:
            # first activate console
            cls.update(id, memory=None, cores=None, console=True,
                       password=None, background=False, max_memory=None)
        # now we can connect
        # retrieve ip of vm
        vm_info = cls.info(id)
        version, ip_addr = cls.vm_ip(id)

        console_url = vm_info.get('console_url', 'console.gandi.net')
        access = 'ssh %s@%s' % (ip_addr, console_url)
        cls.execute(access)


class Image(GandiModule):

    """ Module to handle CLI commands.

    $ gandi vm images

    """

    @classmethod
    def list(cls, datacenter=None, label=None):
        """List available images for vm creation."""
        options = {}
        if datacenter:
            datacenter_id = int(Datacenter.usable_id(datacenter))
            options['datacenter_id'] = datacenter_id

        # implement a filter by label as API doesn't handle it
        images = cls.safe_call('hosting.image.list', options)
        if not label:
            return images
        return [img for img in images
                if label.lower() in img['label'].lower()]

    @classmethod
    def is_deprecated(cls, label, datacenter=None):
        """Check if image if flagged as deprecated."""
        images = cls.list(datacenter, label)
        images_visibility = dict([(image['label'], image['visibility'])
                                  for image in images])
        return images_visibility.get(label, 'all') == 'deprecated'

    @classmethod
    def from_label(cls, label, datacenter=None):
        """Retrieve disk image id associated to a label."""
        result = cls.list(datacenter=datacenter)
        image_labels = dict([(image['label'], image['disk_id'])
                             for image in result])

        return image_labels.get(label)

    @classmethod
    def from_sysdisk(cls, label):
        """Retrieve disk id from available system disks"""
        disks = cls.safe_call('hosting.disk.list', {'name': label})
        if len(disks):
            return disks[0]['id']

    @classmethod
    def usable_id(cls, id, datacenter=None):
        """ Retrieve id from input which can be label or id."""
        try:
            qry_id = int(id)
        except Exception:
            # if id is a string, prefer a system disk then a label
            qry_id = cls.from_sysdisk(id) or cls.from_label(id, datacenter)

        if not qry_id:
            msg = 'unknown identifier %s' % id
            cls.error(msg)

        return qry_id


class Kernel(GandiModule):

    """ Module to handle Gandi Kernels. """

    @classmethod
    def list(cls, datacenter=None, flavor=None, match='', exact_match=False):
        """ List available kernels for datacenter."""
        if not datacenter:
            dc_ids = [dc['id'] for dc in Datacenter.filtered_list()]
            kmap = {}
            for dc_id in dc_ids:
                vals = cls.safe_call('hosting.disk.list_kernels', dc_id)
                for key in vals:
                    kmap.setdefault(key, []).extend(vals.get(key, []))
            # remove duplicates
            for key in kmap:
                kmap[key] = list(set(kmap[key]))
        else:
            dc_id = Datacenter.usable_id(datacenter)
            kmap = cls.safe_call('hosting.disk.list_kernels', dc_id)

        if match:
            for flav in kmap:
                if exact_match:
                    kmap[flav] = [x for x in kmap[flav] if match == x]
                else:
                    kmap[flav] = [x for x in kmap[flav] if match in x]
        if flavor:
            if flavor not in kmap:
                cls.error('flavor %s not supported here' % flavor)
            return dict([(flavor, kmap[flavor])])

        return kmap

    @classmethod
    def is_available(cls, disk, kernel):
        """ Check if kernel is available for disk."""
        kmap = cls.list(disk['datacenter_id'], None, kernel, True)
        for flavor in kmap:
            if kernel in kmap[flavor]:
                return True
        return False



""" Domain namespace commands. """

import click

# define unicode for python3
try:
    unicode
except NameError:
    unicode = str

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import output_contact_info, output_domain
from gandi.cli.core.params import pass_gandi


@cli.group(name='domain')
@pass_gandi
def domain(gandi):
    """Commands related to domains."""


@domain.command()
@click.option('--limit', help='Limit number of results.', default=100,
              show_default=True)
@pass_gandi
def list(gandi, limit):
    """List domains."""
    options = {'items_per_page': limit}
    domains = gandi.domain.list(options)
    for domain in domains:
        gandi.echo(domain['fqdn'])

    return domains


@domain.command()
@click.argument('resource')
@pass_gandi
def info(gandi, resource):
    """Display information about a domain."""
    output_keys = ['fqdn', 'nameservers', 'services', 'zone_id', 'tags',
                   'created', 'expires', 'updated']
    contact_field = ['owner', 'admin', 'bill', 'tech', 'reseller']

    result = gandi.domain.info(resource)
    output_contact_info(gandi, result['contacts'], contact_field, justify=12)
    output_domain(gandi, result, output_keys, justify=12)

    return result


@domain.command()
@click.option('--domain', default=None, help='Name of the domain.')
@click.option('--duration', default=1, prompt=True,
              type=click.IntRange(min=1, max=10),
              help='Registration period in years, between 1 and 10.')
@click.option('--owner', default=None,
              help='Registrant handle.')
@click.option('--admin', default=None,
              help='Administrative contact handle.')
@click.option('--tech', default=None,
              help='Technical contact handle.')
@click.option('--bill', default=None,
              help='Billing contact handle.')
@click.option('--nameserver', default=None,
              help='Nameserver', multiple=True)
@click.option('--extra_parameter', default=None, nargs=2,
              type=(unicode, unicode),
              help='Extra parameters', multiple=True)
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.argument('resource', metavar='DOMAIN', required=False)
@pass_gandi
def create(gandi, resource, domain, duration, owner, admin, tech, bill,
           nameserver, extra_parameter, background):
    """Buy a domain."""
    if domain:
        gandi.echo('/!\ --domain option is deprecated and will be removed '
                   'upon next release.')
        gandi.echo("You should use 'gandi domain create %s' instead." % domain)

    if (domain and resource) and (domain != resource):
        gandi.echo('/!\ You specified both an option and an argument which '
                   'are different, please choose only one between: %s and %s.'
                   % (domain, resource))
        return

    _domain = domain or resource
    if not _domain:
        _domain = click.prompt('Name of the domain')

    result = gandi.domain.create(_domain, duration, owner, admin, tech, bill,
                                 nameserver, extra_parameter, background)
    if background:
        gandi.pretty_echo(result)

    return result


@domain.command()
@click.option('--duration', default=1, prompt=True,
              type=click.IntRange(min=1, max=10),
              help='Registration period in years, between 1 and 10.')
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.argument('domain')
@pass_gandi
def renew(gandi, domain, duration, background):
    """Renew a domain."""
    result = gandi.domain.renew(domain, duration, background)
    if background:
        gandi.pretty_echo(result)

    return result



""" Certificate namespace commands. """

import os
import click
import requests

# define basestring for python3
try:
    basestring
except NameError:
    basestring = (str, bytes)

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import output_cert, output_cert_oper, display_rows
from gandi.cli.core.params import (pass_gandi, IntChoice,
                                   CERTIFICATE_PACKAGE, CERTIFICATE_DCV_METHOD,
                                   CERTIFICATE_PACKAGE_TYPE,
                                   CERTIFICATE_PACKAGE_MAX,
                                   CERTIFICATE_PACKAGE_WARRANTY)

type_list = list


@cli.group(name='certificate')
@pass_gandi
def certificate(gandi):
    """Commands related to certificates."""


@certificate.command()
@pass_gandi
def packages(gandi):
    """ List certificate packages.
    /!\\ deprecated call.
    """
    gandi.echo('/!\ "gandi certificate packages" is deprecated.')
    gandi.echo('Please use "gandi certificate plans".')
    return _plans(gandi, with_name=True)


@certificate.command()
@pass_gandi
def plans(gandi):
    """ List certificate plans. """
    return _plans(gandi)


def package_desc(gandi, package):
    if isinstance(package, basestring):
        package = gandi.certificate.package_get(package)
        if not package:
            return ''

    type_ = package['category']['name']
    if package['wildcard']:
        desc = '%s wildcard' % type_
    elif package['max_domains'] > 1:
        desc = '%s multi domain' % type_
    else:
        desc = '%s single domain' % type_
    return ' '.join([word.capitalize() for word in desc.split(' ')])


def _plans(gandi, with_name=False):
    packages = gandi.certificate.package_list()

    def keyfunc(item):
        return (item['category']['id'],
                item['max_domains'],
                item['warranty'],
                item['name'])

    packages.sort(key=keyfunc)
    labels = ['Description', 'Max altnames', 'Type']
    if with_name:
        labels.insert(1, 'Name')
    else:
        labels.append('Warranty')
    ret = [labels]

    for package in packages:
        cat = package['name'].split('_')[1]
        warranty = str(int(package['name'].split('_')[3]) * 1000)
        if len(warranty) > 3:
            warranty = type_list(warranty)
            warranty.insert(len(warranty) - 3, ',')
            warranty = ''.join(warranty)

        desc = package_desc(gandi, package)
        line = [desc, str(package['max_domains']), cat]
        if with_name:
            line.insert(1, package['name'])
        else:
            line.append(warranty)
        ret.append(line)

    display_rows(gandi, ret)

    return ret


@certificate.command()
@click.option('--id', help='Display ids.', is_flag=True)
@click.option('--altnames', help='Display altnames.', is_flag=True)
@click.option('--csr', help='Display CSR.', is_flag=True)
@click.option('--cert', help='Display CRT.', is_flag=True)
@click.option('--all-status', is_flag=True,
              help='Filter the certificate without regard to its status.')
@click.option('--status', help='Display status.', is_flag=True)
@click.option('--dates', help='Display dates.', is_flag=True)
@click.option('--limit', help='Limit number of results.', default=100,
              show_default=True)
@pass_gandi
def list(gandi, id, altnames, csr, cert, all_status, status, dates, limit):
    """ List certificates. """
    options = {'items_per_page': limit}

    if not all_status:
        options['status'] = ['valid', 'pending']

    output_keys = ['cn', 'plan']

    if id:
        output_keys.append('id')

    if status:
        output_keys.append('status')

    if dates:
        output_keys.extend(['date_created', 'date_end'])

    if altnames:
        output_keys.append('altnames')

    if csr:
        output_keys.append('csr')

    if cert:
        output_keys.append('cert')

    result = gandi.certificate.list(options)
    for num, cert in enumerate(result):
        if num:
            gandi.separator_line()
        cert['plan'] = package_desc(gandi, cert['package'])
        output_cert(gandi, cert, output_keys)

    return result


@certificate.command()
@click.argument('resource', nargs=-1, required=True)
@click.option('--id', help='Display ids.', is_flag=True)
@click.option('--altnames', help='Display altnames.', is_flag=True)
@click.option('--csr', help='Display CSR.', is_flag=True)
@click.option('--cert', help='Display CRT.', is_flag=True)
@click.option('--all-status', help='Show all certificates.', is_flag=True)
@pass_gandi
def info(gandi, resource, id, altnames, csr, cert, all_status):
    """ Display information about a certificate.

    Resource can be a CN or an ID
    """
    output_keys = ['cn', 'date_created', 'date_end', 'plan', 'status']

    if id:
        output_keys.append('id')

    if altnames:
        output_keys.append('altnames')

    if csr:
        output_keys.append('csr')

    if cert:
        output_keys.append('cert')

    ids = []
    for res in resource:
        ids.extend(gandi.certificate.usable_ids(res))

    result = []
    for num, id_ in enumerate(set(ids)):
        cert = gandi.certificate.info(id_)
        if not all_status and cert['status'] not in ['valid', 'pending']:
            continue
        if num:
            gandi.separator_line()
        cert['plan'] = package_desc(gandi, cert['package'])
        output_cert(gandi, cert, output_keys)
        result.append(cert)

    return result


@certificate.command()
@click.argument('resource', nargs=-1, required=True)
@click.option('-o', '--output', help='The file to write the cert.')
@click.option('--force', '-f', is_flag=True,
              help='Overwrite the crt file if it exists.')
@click.option('-i', '--intermediate', is_flag=True,
              help='Retrieve gandi intermediate certs.')
@pass_gandi
def export(gandi, resource, output, force, intermediate):
    """ Write the certificate to <output> or <fqdn>.crt.

    Resource can be a CN or an ID
    """
    ids = []
    for res in resource:
        ids.extend(gandi.certificate.usable_ids(res))

    if output and len(ids) > 1:
        gandi.echo('Too many certs found, you must specify which cert you '
                   'want to export')
        return

    for id_ in set(ids):
        cert = gandi.certificate.info(id_)
        if 'cert' not in cert:
            continue

        if cert['status'] != 'valid':
            gandi.echo('The certificate must be in valid status to be '
                       'exported (%s).' % id_)
            continue

        cert_filename = cert['cn'].replace('*.', 'wildcard.', 1)
        crt_filename = output or cert_filename + '.crt'
        if not force and os.path.isfile(crt_filename):
            gandi.echo('The file %s already exists.' % crt_filename)
            continue

        crt = gandi.certificate.pretty_format_cert(cert)
        if crt:
            with open(crt_filename, 'w') as crt_file:
                crt_file.write(crt)
                gandi.echo('wrote %s' % crt_filename)

        package = cert['package']
        if 'bus' in package and intermediate:
            gandi.echo('Business certs do not need intermediates.')
        elif intermediate:
            crtf = 'pem'
            sha_version = cert['sha_version']
            type_ = package.split('_')[1]
            extra = ('sgc' if 'SGC' in package
                     and 'pro' in package
                     and sha_version == 1 else 'default')

            if extra == 'sgc':
                crtf = 'pem'

            inters = gandi.certificate.urls[sha_version][type_][extra][crtf]
            if isinstance(inters, basestring):
                inters = [inters]

            fhandle = open(cert_filename + '.inter.crt', 'w+b')
            for inter in inters:
                if inter.startswith('http'):
                    data = requests.get(inter).text
                else:
                    data = inter
                fhandle.write(data.encode('latin1'))

            gandi.echo('wrote %s' % cert_filename + '.inter.crt')
            fhandle.close()

        return crt


@certificate.command()
@click.option('--csr', required=False,
              help='Csr of the new certificate (filename or content).')
@click.option('--private-key', '--pk', required=False,
              help='Private key to use to generate the CSR (filename or '
              'content).')
@click.option('--common-name', '--cn', required=False,
              help='Common name to use when generating the CSR.')
@click.option('--country', required=False,
              help='The generated CSR country (C).')
@click.option('--state', required=False,
              help='The generated CSR state (ST).')
@click.option('--city', required=False,
              help='The generated CSR location (L).')
@click.option('--organisation', required=False,
              help='The generated CSR organisation (O).')
@click.option('--branch', required=False,
              help='The generated CSR branch (OU).')
@click.option('-d', '--duration', default=1,
              type=IntChoice(['1', '2']),
              help='The certificate duration in year.')
@click.option('--package', type=CERTIFICATE_PACKAGE,
              help='Certificate package.')
@click.option('--type', type=CERTIFICATE_PACKAGE_TYPE,
              help='Certificate package type (default=std).')
@click.option('--max-altname', type=CERTIFICATE_PACKAGE_MAX,
              help='Certificate package max altname number.')
@click.option('--warranty', type=CERTIFICATE_PACKAGE_WARRANTY,
              help='Certificate warranty, only good for pro certificates.')
@click.option('--altnames', required=False, multiple=True,
              help='The certificate altnames (comma separated text without '
                   'space).')
@click.option('--dcv-method', required=False, type=CERTIFICATE_DCV_METHOD,
              help='Give the DCV method to use to check domain ownership.')
@pass_gandi
def create(gandi, csr, private_key, common_name, country, state, city,
           organisation, branch, duration, package, type, max_altname,
           warranty, altnames, dcv_method):
    """Create a new certificate."""
    if not (csr or common_name):
        gandi.echo('You need a CSR or a CN to create a certificate.')
        return

    if package and (type or max_altname or warranty):
        gandi.echo('Please do not use --package at the same time you use '
                   '--type, --max-altname or --warranty.')
        return

    if type and warranty and type != 'pro':
        gandi.echo('The warranty can only be specified for pro certificates.')
        return

    csr = gandi.certificate.process_csr(common_name, csr, private_key, country,
                                        state, city, organisation, branch)
    if not csr:
        return

    if not common_name:
        common_name = gandi.certificate.get_common_name(csr)
        if not common_name:
            gandi.echo('Unable to parse provided csr: %s' % csr)
            return

    if '*' in common_name and altnames and len(altnames) > 1:
        gandi.echo("You can't have a wildcard with multidomain certificate.")
        return

    if package:
        gandi.echo('/!\ Using --package is deprecated, please replace it by '
                   '--type (in std, pro or bus) and --max-altname to set '
                   'the max number of altnames.')
    elif type or max_altname or warranty:
        package = gandi.certificate.get_package(common_name,
                                                type,
                                                max_altname,
                                                altnames,
                                                warranty)
        if not package:
            gandi.echo("Can't find any plan with your params.")
            gandi.echo('Please call : "gandi certificate plans".')
            return

    result = gandi.certificate.create(csr, duration, package, altnames,
                                      dcv_method)

    gandi.echo('The certificate create operation is %s' % result['id'])
    gandi.echo('You can follow it with:')
    gandi.echo('$ gandi certificate follow %s' % result['id'])
    if common_name:
        gandi.echo('When the operation is DONE, you can retrieve the .crt'
                   ' with:')
        gandi.echo('$ gandi certificate export "%s"' % common_name)

    return result


@certificate.command()
@click.argument('resource', nargs=1, required=True)
@click.option('--csr', help='New csr for the certificate.', required=False)
@click.option('--private-key', '--pk', required=False,
              help='Private key to use to generate the CSR.')
@click.option('--country', '--c', required=False,
              help='The generated CSR country (C).')
@click.option('--state', '--st', required=False,
              help='The generated CSR state (ST).')
@click.option('--city', '--l', required=False,
              help='The generated CSR location (L).')
@click.option('--organisation', '--o', required=False,
              help='The generated CSR organisation (O).')
@click.option('--branch', '--ou', required=False,
              help='The generated CSR branch (OU).')
@click.option('--altnames', required=False, multiple=True,
              help='All the certificate altnames (comma separated text '
                   'without space).')
@click.option('--dcv-method', required=False, type=CERTIFICATE_DCV_METHOD,
              help='Give the DCV method to use to check domain ownership.')
@pass_gandi
def update(gandi, resource, csr, private_key, country, state, city,
           organisation, branch, altnames, dcv_method):
    """ Update a certificate CSR.

    Resource can be a CN or an ID
    """
    ids = gandi.certificate.usable_ids(resource)

    if len(ids) > 1:
        gandi.echo('Will not update, %s is not precise enough.' % resource)
        gandi.echo('  * cert : ' +
                   '\n  * cert : '.join([str(id_) for id_ in ids]))
        return

    id_ = ids[0]

    result = gandi.certificate.update(id_, csr, private_key, country, state,
                                      city, organisation, branch, altnames,
                                      dcv_method)

    gandi.echo('The certificate update operation is %s' % result['id'])
    gandi.echo('You can follow it with:')
    gandi.echo('$ gandi certificate follow %s' % result['id'])
    gandi.echo('When the operation is DONE, you can retrieve the .crt'
               ' with:')
    gandi.echo('$ gandi certificate export "%s"' % resource)

    return result


@certificate.command()
@click.argument('resource', nargs=1, required=True)
@pass_gandi
def follow(gandi, resource):
    """ Get the operation status

    Resource is an operation ID
    """
    oper = gandi.oper.info(int(resource))
    assert(oper['type'].startswith('certificate_'))
    output_cert_oper(gandi, oper)
    return oper


@certificate.command('change-dcv')
@click.argument('resource', nargs=1, required=True)
@click.option('--dcv-method', required=True, type=CERTIFICATE_DCV_METHOD,
              help='Give the updated DCV method to use.')
@pass_gandi
def change_dcv(gandi, resource, dcv_method):
    """ Change the DCV for a running certificate operation.

    Resource can be a CN or an ID
    """
    ids = gandi.certificate.usable_ids(resource)

    if len(ids) > 1:
        gandi.echo('Will not update, %s is not precise enough.' % resource)
        gandi.echo('  * cert : ' +
                   '\n  * cert : '.join([str(id_) for id_ in ids]))
        return

    id_ = ids[0]

    opers = gandi.oper.list({'cert_id': id_})
    if not opers:
        gandi.echo('Can not find any operation for this certificate.')
        return

    oper = opers[0]
    if (oper['step'] != 'RUN'
            and oper['params']['inner_step'] != 'comodo_oper_updated'):
        gandi.echo('This certificate operation is not in the good step to '
                   'update the DCV method.')
        return

    gandi.certificate.change_dcv(oper['id'], dcv_method)
    cert = gandi.certificate.info(id_)

    csr = oper['params']['csr']
    package = cert['package']
    altnames = oper['params'].get('altnames')
    gandi.certificate.advice_dcv_method(csr, package, altnames, dcv_method,
                                        cert_id=id_)


@certificate.command('resend-dcv')
@click.argument('resource', nargs=1, required=True)
@pass_gandi
def resend_dcv(gandi, resource):
    """ Resend the DCV mail.

    Resource can be a CN or an ID
    """
    ids = gandi.certificate.usable_ids(resource)

    if len(ids) > 1:
        gandi.echo('Will not update, %s is not precise enough.' % resource)
        gandi.echo('  * cert : ' +
                   '\n  * cert : '.join([str(id_) for id_ in ids]))
        return

    id_ = ids[0]

    opers = gandi.oper.list({'cert_id': id_})
    if not opers:
        gandi.echo('Can not find any operation for this certificate.')
        return

    oper = opers[0]
    if (oper['step'] != 'RUN'
            and oper['params']['inner_step'] != 'comodo_oper_updated'):
        gandi.echo('This certificate operation is not in the good step to '
                   'resend the DCV.')
        return

    if oper['params']['dcv_method'] != 'email':
        gandi.echo('This certificate operation is not in email DCV.')
        return

    gandi.certificate.resend_dcv(oper['id'])


@certificate.command()
@click.argument('resource', nargs=1, required=True)
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@pass_gandi
def delete(gandi, resource, background, force):
    """ Revoke the certificate.

    Resource can be a CN or an ID
    """
    ids = gandi.certificate.usable_ids(resource)

    if len(ids) > 1:
        gandi.echo('Will not delete, %s is not precise enough.' % resource)
        gandi.echo('  * cert : ' +
                   '\n  * cert : '.join([str(id_) for id_ in ids]))
        return

    if not force:
        proceed = click.confirm("Are you sure to delete the certificate %s?" %
                                resource)
        if not proceed:
            return

    result = gandi.certificate.delete(ids[0], background)
    return result
